/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/*************************************************************************
 *
 *   Name     : fft.c
 *   Version  : 1999/08/04 7.0
 *   Comment  : Fast-Fourier-Transform (FFT) and inverse FFT
 *   Author   : Markus Mummert
 *   Reference:	Hess, W., "Spektrale Methoden der Signalverarbeitung",
 *		accompanying material to a lecture held at
 *		Technische Universitaet Muenchen, summer-term 1985
 *   History  :
 *   1999/08/04 7.0 allow multiple use - no persistent internal data; 
 *		removed FTTMAXDIM-check; changed user interface !
 *   19.10.98/6 renamed exp2 -> intexp2; void-declarations
 *   11.10.98/5 extern-declaration malloc() -> portab.h
 *   15.8.98/4  revert to manual history
 *   revision 2.0 1993/07/01 01:52:42 mummert
 *		saving multiplications by separately handling FFT-runs 0-3
 *   revision 1.1 1993/06/29 21:52:27 mummert
 *		improved interface
 *   revision 1.0 1993/06/29 18:43:35 mummert
 *		functional
 * 
 */

#include <stdio.h>
#include <math.h>
/* #include "portab.h" should be included via fft.h */
#include "fft.h"

/*
 * intexp2 - 2**x for integer x
 *
 */
static int intexp2(x)
register int x;
{
	register int y = 1;

	for (; x-- > 0; )
		y*=2;
	return(y);
}

/*
 * revtab - set up address table for bit reversal
 *
 *	irev	addressing table
 *	nrun
 *	lrun
 *
 */
static void revtab(irev, nrun, lrun)
register int *irev;
int nrun, lrun;
{
	int i, j, k, m;

	for (i = 0; i <= nrun; i++) {
		k = i;
		irev[i] = 0;
		for (j = 0; j <= lrun; j++) {
			m = k % 2;
			k = k / 2;
			irev[i] = irev[i]*2 + m;
		}
	}
}

/*
 * bitrev - do bit reversal
 *
 *	irev	addressing table
 *	ldn	log2 of dimension
 *	xr,xi	real/imaginary-array for reversal
 *
 */
static void bitrev(irev, ldn, xr, xi)
int ldn, *irev;
register FFTELEM *xr, *xi;
{
	register FFTELEM z;
	register int i, j;
	int n1;

	n1 = intexp2(ldn) - 1;

	for (i = 0; i <= n1; i++) {
		j = irev[i];
		if (i < j) {
			z = xr[i];
			xr[i] = xr[j];
			xr[j] = z;
			z = xi[i];
			xi[i] = xi[j];
			xi[j] = z;
		}
	}
}

/*
 * tab - make tables
 *
 *	irev	addressing table
 *	ldn	log2 of dimension
 *	co,si	cosine/sine-table (bit-reversed)
 */
static void tab(irev, ldn, co, si)
int *irev, ldn;
FFTELEM *co, *si;
{
	int i, n, n1, ldn1, nh, nh1, nrun, lrun;
	double a, d;

	ldn1 = ldn - 1;
	nh = intexp2(ldn1);
	n = 2 * nh;
	n1 = n - 1;
	nh1 = nh - 1;

	a = -2*M_PI/n;
	for (i = 0; i<= nh1; i++) {
		d = a * i;
		si[i] = sin(d);
		co[i] = cos(d);
	}
	nrun = nh1;
	lrun = ldn1 - 1;
	revtab(irev, nrun, lrun);

	bitrev(irev, ldn1, co, si);

	nrun = n1;
	lrun = ldn1;
	revtab(irev, nrun, lrun);
}

/*
 * fft_init - allocate and initialize FFT-arrays
 *
 *	ldn	log2 of dimension
 *
 *	return pointer to FTTDATA-struct (handle)
 *      or NULL if not enough memory
 */
FFTDATA *fft_init(ldn)
int ldn;
{
	int n, *irev;
	FFTDATA *datap;
	FFTELEM *CO, *SI, *XI, *XR;

	n = intexp2(ldn);

	datap = (FFTDATA *)(malloc(sizeof(FFTDATA)));
	XI    = (FFTELEM *)(malloc(sizeof(FFTELEM)*n));
	XR    = (FFTELEM *)(malloc(sizeof(FFTELEM)*n));
	CO    = (FFTELEM *)(malloc(sizeof(FFTELEM)*n));
	SI    = (FFTELEM *)(malloc(sizeof(FFTELEM)*n));
	irev  = (int *)(malloc(sizeof(int)*n));

	if (!(XI && XR && CO && SI && irev && datap))
		return(NULL);

	tab(irev, ldn, CO, SI);
	datap->r    = XR;
	datap->i    = XI;
	datap->co   = CO;
	datap->si   = SI;
	datap->irev = irev;
	datap->ldn  = ldn;
	datap->n    = n;

	return(datap);
}

/*
 * fft_discard - discard FFT-memory
 *
 *	datap	handle
 */
void fft_discard(datap)
FFTDATA *datap;
{
	free(datap->r);	
	free(datap->i);	
	free(datap->co);	
	free(datap->si);	
	free(datap->irev);	
	free(datap);
}

/*
 * fft - Fast Fourier Transform
 *
 *	datap	handle
 */
void fft(datap)
FFTDATA *datap;
{
	FFTELEM sqr, zr, zi;
	register FFTELEM *CO, *SI, *XI, *XR;
	register int k, i;
	int ldn, *irev, iw, ibegin, iend, istep, ni, n1, nk, ldn1, selector;

	XR = datap->r;
	XI = datap->i;
	CO = datap->co;
	SI = datap->si;
	irev = datap->irev;
	ldn  = datap->ldn;

	n1 = intexp2(ldn) - 1;
	ldn1 = ldn - 1;

	for (istep = 0; istep <= ldn1; istep++) {
		nk = intexp2(ldn1 - istep);
		ni = nk + nk;
		iw = 0;
		selector = 0;
		for (ibegin = 0; ibegin <= n1; ibegin += ni) {
			iend = ibegin + nk - 1;
			switch(selector++) {
			case 0:
				for (i = ibegin; i <= iend; i++) {
					k = i + nk;
					zr = XR[k];
					zi = XI[k];
					XR[k] = XR[i] - zr;
					XI[k] = XI[i] - zi;
					XR[i] += zr;
					XI[i] += zi;
				}
				break;
			case 1:
				for (i = ibegin; i <= iend; i++) {
					k = i + nk;
					zr =   XI[k];
					zi = - XR[k];
					XR[k] = XR[i] - zr;
					XI[k] = XI[i] - zi;
					XR[i] += zr;
					XI[i] += zi;
				}
				break;
			case 2:
				sqr = CO[2];
				for (i = ibegin; i <= iend; i++) {
					k = i + nk;
					zr = (  XR[k] + XI[k])*sqr;
					zi = (- XR[k] + XI[k])*sqr;
					XR[k] = XR[i] - zr;
					XI[k] = XI[i] - zi;
					XR[i] += zr;
					XI[i] += zi;
				}
				break;
			case 3:
				for (i = ibegin; i <= iend; i++) {
					k = i + nk;
					zr =   (- XR[k] + XI[k])*sqr;
					zi = - (  XR[k] + XI[k])*sqr;
					XR[k] = XR[i] - zr;
					XI[k] = XI[i] - zi;
					XR[i] += zr;
					XI[i] += zi;
				}
				break;
			default:
				for (i = ibegin; i <= iend; i++) {
					k = i + nk;
					zr = XR[k]*CO[iw] - XI[k]*SI[iw];
					zi = XR[k]*SI[iw] + XI[k]*CO[iw];
					XR[k] = XR[i] - zr;
					XI[k] = XI[i] - zi;
					XR[i] += zr;
					XI[i] += zi;
				}
				break;
			}
			iw++;
		}
	}
	bitrev(irev, ldn, XR, XI); 
}

/*
 * ifft - inverse FFT
 *
 *	datap	handle
 */
void ifft(datap)
FFTDATA *datap;
{
	register int i, n;
	FFTELEM invn, m1;
	register FFTELEM *XI, *XR;

	XR = datap->r;
	XI = datap->i;
	n = datap->n;
	invn = 1.0/n;
	m1 = -1.0;

	/* conjugate */
	for (i = 0; i < n; i++) {
		XI[i]*=m1;
	}
	fft(datap);
	/* conjugate, scale with 1/n */
	for (i = 0; i < n; i++) {
		XR[i]*=invn;
		XI[i]*=-invn;
	}
}
/*
 * EOT
 *
 ************************************************************************/
