/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/************************************************************************/
/*									*/
/*  Datei    :	KEY.C							*/
/*  Version  :	1999/02/22 - 10						*/
/*  Autor    :	Markus Mummert						*/
/*									*/
/*  Kommentar:	Modul zum Management von Eingabeparametern bei komman-	*/
/*		dozeilenorientierten Programmen. Grundkonzept ist die	*/
/*		Zuordnung von Parametern aus Kommandozeile oder Datei	*/
/*		durch vorangestellte Schluesselwoerter (Keywords).	*/
/*									*/
/*		Das Hauptmodul muss zum einen eine Tabelle, genannt	*/
/*		Keytab (s.a. key.h), definieren. Dort wird die Zuord-	*/
/*		nung von Keywords zu Programmvariablen festgelegt,nebst	*/
/*		Textkommentar, internem Default-Wert und zulaessigem	*/
/*		Eingabebereich.	Weiterhin muss die Funktion header()	*/
/*		bereitgestellt werden, die im Zusammenwirken mit der	*/
/*		Hilfstextausgabe (s.a. HelpText()) einen Beitrag des	*/
/*		Hauptmoduls leistet.					*/
/*									*/
/*		Der Aufruf der Funktion initkeytab() durch das Haupt-	*/
/*		modul bewirkt den automatischen Zugriff auf Parmeter-	*/
/*		eingabequellen mit Zuweisung der Eingaben per Keywords	*/
/*		an die in der Keytab zugeordneten Variablen. Quellen	*/
/*		sind Kommandozeile, optionale Parameterdatei, Default-	*/
/*		Parameterdatei und die internen Keytab-Default-Werte	*/
/*		(fallende Prioritaet). Auch Zusatzfunktionen wie	*/
/*		Anzeige oder Speichern der aktuellen Parameter,		*/
/*		Hilfstextausgabe, Bereichskontrolle und Fehlermeldung	*/
/*		werden autonom durchgefuehrt.				*/
/*									*/
/*		Die Funktionen catmsg() und errorhandler()		*/
/*		erlauben dem Hauptmodul Fehlermeldungen ohne weitere	*/
/*		Ausgaberoutinen abzusetzen. catmsg() verbindet zwei	*/
/*		Text-Strings miteinander und ermoeglicht den sukzes-	*/
/*		siven Aufbau von Fehlermeldungen ueber mehrere		*/
/*		Unterprogrammstufen, die dann per errorhandler()	*/
/*		ausgebeben werden koennen.				*/
/*									*/
/*		Die Funktion convspchar() kann benutzt werden, um in	*/
/*		Strings Sonderzeichen in Form von 'C'-Escape-Sequenzen	*/
/*		umzuwandeln die entsprechenden 'echten' Sonderzeichen.	*/
/*									*/
/*		Weitere Informationen in key.h (anwenderorientiert)	*/
/*									*/
/*  history:	Erstversion Anlehnung an KEYTAB.C von U.Baumann		*/
/*		7.9.90/3						*/
/*		    Lauffaehige Version					*/
/*		16.11.90/4						*/
/*		    In StreamToKeytab() char *p nun static;		*/
/*		    Strings mit Leerzeichen erscheinen in "";		*/
/*		    Zulassungsmuster fuer Strings auch Leerzeichen;	*/
/*		1.3.93/5						*/
/*		    Helptext() auf stderr, INT mit %12d			*/
/*		19.6.98/6						*/
/*		    Neuaufbau Funktionalitaet von 16.11.90/4		*/
/*		    auf Basis von 7.9.90/3 wegen diverser Fehler;	*/
/*		    neue Option NOLIMITS schaltet CheckInputRange() ab;	*/
/*		    kein Ueberschreiben von String-Konstanten mehr;	*/
/*		    malloc() statt fix-Buffer in StreamToKeytab();	*/
/*		    kein Abschneiden bei Platzmangel in INFO-Ausgabe;	*/
/*		    Helptext() wird auf stderr ausgegeben		*/
/*		26.6.98/7						*/
/*		    Kommentar bei INFO springt in naechste Zeile, wenn 	*/
/*		    bei STRING Variablenwert ueberlang  	  	*/
/*		28.6.98/8						*/
/*		    Mehrere LOADPARM-Optionen moeglich; Helptext verb. 	*/
/*		27.8.98/9						*/
/*		    DFLTPARMPATH neu; includes von .h nach .c;		*/
/*		    portab.h eingefuegt					*/
/*		1999/02/22 - 10						*/
/*		    All user messages in English			*/
/*									*/
/************************************************************************/

#define	KEY_SOURCE			/* Source-Identifikation	*/

#include <stdio.h>
#include <string.h>
#include "portab.h"
#include "key.h"			/* Import/Export-Datei		*/

#define LOADPARMOPT	"LOADPARM"	/* Optionsnamen, s. HelpText()	*/
#define SAVEPARMOPT	"SAVEPARM"
#define SHOWPARMOPT	"INFOPARM"
#define FLASHPARMOPT	"INFO"
#define INTNDFTSOPT	"DEFAULTS"
#define NOLIMOPT	"NOLIMITS"
#define MAXLOAD		256		/* max. Anz. LOADPARM-Optionen	*/

#ifndef DFLTPARMPATH
#define DFLTPARMPATH	"."		/* Pfad Defaultparameterfile	*/
#endif

static Helptext();			/* static-Deklarationen		*/
static int SetVarFromKeytab();
static int CmdlineToKeytab();
static int ParmFileToKeytab();
static int KeytabToParmFile();
static LongInfoParm();
static InfoParm();
static int CheckInputRange();
static int FindKtEntry();
static int StringCompare1();
static int StringCompare2();
static int StreamToKeytab();
static int KeytabToStream();
static int NextKwdKargStream();
static int g_nolim = -1;		/* >= 0 -> kein CheckInputRange	*/
static char *g_myname;

/*----------------------------------------------------------------------*/
/*  initkeytab	besetzt die in der Keytab KTP angegeben Variablen mit	*/
/*		Werten aus folgenden Quellen, wobei die Prioritaetstufe	*/
/*		in der Aufstellung zunimmt:				*/
/*									*/
/*		1. Keytab-interne Default-Werte				*/
/*		2. Werte aus Default-Parameterdatei, wenn vorhanden	*/
/*		3. Werte aus Parameterdatei, wenn diese Option in	*/
/*		   Kommandozeile spezifiziert wurde.			*/
/*		4. Werte aus Kommandozeile				*/
/*									*/
/*		Die Zuordnung von Werten aus Datei oder Kommandozeile	*/
/*		erfolgt ueber die vorangestellten Keywords der Keytab.	*/
/*		Name zu 2. durch DFTPFNAME uebergeben.			*/
/*									*/
/*		ARGC, ARGV sind die Funktionsargumente von main().	*/
/*		Return != 0 als Abbruchkennung.				*/
/*----------------------------------------------------------------------*/
int initkeytab(argc, argv, ktp, dftpfname)
int argc;
char *argv[], *dftpfname;
KEY_TAB *ktp;
{
	static int load[MAXLOAD];
	int i, nl = 0, save, show, dfts, flash;
	char *errmsg, defaultfn[MAXLINE];

	g_myname = argv[0];
	argc--; argv++;			/* Programmaufruf ausblenden	*/
	if (argc == 0) {		/* Hilfstext ausgeben		*/
		Helptext(dftpfname);
		return(-1);
	} else {
					/* Keytab-Variablen auf interne */
					/* Default-Keytab-Werte setzen	*/
 		if (SetVarFromKeytab(ktp, &errmsg)) {
			errmsg = catmsg("in keytab", errmsg);
			errorhandler(errmsg);
			return(-1);
		}			/* zunaechst nur Kommandozeilen-*/
					/* Syntaxueberpruefung wichtig,	*/
					/* sowie Feststellung der Opt.	*/
		if (CmdlineToKeytab(argc, argv, ktp,
		    &show, &flash, load, &nl, &save, &dfts, &errmsg)){
			errmsg = catmsg("on command line", errmsg);
			errorhandler(errmsg);
			return(-1);
		}			/* Keytab-Werte-Zg. auf Werte	*/
					/* aus Default-Parameterdatei	*/
					/* umbiegen, wenn diese vorh.	*/
		sprintf(defaultfn, "%s/%s", DFLTPARMPATH, dftpfname);
		if (ParmFileToKeytab(defaultfn, ktp, &errmsg) < 0) {
			errmsg =
			    catmsg("reading default-parameter file", errmsg);
			errorhandler(errmsg);
			return(-1);	/* (Fehlerbeachtung nur wenn da)*/
		}
					/* Option: Keytab-Werte-Zg. auf	*/
					/* Werte aus Parameterdatei	*/
					/* umbiegen			*/
		for (i = 0; i < nl; i++) {
		    if (ParmFileToKeytab(argv[load[i]+1], ktp, &errmsg)) {
			errmsg = catmsg("reading parameter file", errmsg);
			errorhandler(errmsg);
			return(-1);
		    }
		}			/* Keytab-Werte-Zg. auf Werte	*/
					/* aus Kommandozeile umbiegen	*/
		CmdlineToKeytab(argc, argv, ktp,
		    &show, &flash, load, &nl, &save, &dfts, &errmsg);

					/* Option: Keytab zeigen	*/
		if (show >= 0)
			LongInfoParm(ktp);
		else if (flash >= 0)
			InfoParm(ktp);
					/* Option: Keytab Speichern	*/
		if (save >= 0) {
		    if (KeytabToParmFile(argv[save+1], ktp, &errmsg)) {
			errmsg = 
			    catmsg("saving parameter file", errmsg);
			errorhandler(errmsg);
			return(-1);
		    }
		}
	}
	if (show >= 0 || flash >= 0)
		return(-1);		/* Abbruchkennung bei Info	*/
	return(0);
}
/*----------------------------------------------------------------------*/
/*  Helptext		gibt Hilftext auf stdout aus, externe Funktion	*/
/*			header() liefert Beitrag von Hauptmodul.	*/
/*			DFTPFNAME ist der Defaultparameterdateiname	*/
/*----------------------------------------------------------------------*/
static Helptext(dftpfname)
char *dftpfname;
{
char defaultfn[MAXLINE];
header();				/* Beitrag von Hauptmodul	*/
sprintf(defaultfn, "%s/%s", DFLTPARMPATH, dftpfname);
fprintf(stderr,"    Options on command line:\n");
fprintf(stderr,"        %-10s display effective parameters, no calculation\n",
					FLASHPARMOPT);
fprintf(stderr,"        %-10s as above, but long list with parameter ranges\n",
					SHOWPARMOPT);
fprintf(stderr,"        %-10s load parameter file, name follows as argument\n",
					LOADPARMOPT);
fprintf(stderr,"        %-10s save parameter file, name follows as argument\n",
					SAVEPARMOPT);
fprintf(stderr,"        %-10s run on defaults (empty command line shows this text)\n",
					INTNDFTSOPT);
fprintf(stderr,"        %-10s suppress range check of parameters\n\n",
					NOLIMOPT);
fprintf(stderr,"    Parameter sources (ascending order of priority):\n");
fprintf(stderr,"        Internal default parameters\n");
fprintf(stderr,"        Default-parameter file %s\n", defaultfn);
fprintf(stderr,"        Parameter files via %s (ascending priority in line)\n", LOADPARMOPT);
fprintf(stderr,"        Parameters on command line (ascending priority in line)\n");
fprintf(stderr,"\n");
fprintf(stderr,"    Parameters are identified by preceeding keywords. ");
fprintf(stderr,"For legal keywords\n");
fprintf(stderr,"    see %s or %s. ",FLASHPARMOPT, SHOWPARMOPT);
fprintf(stderr,"Keywords are written in capital or small letters,\n");
fprintf(stderr,"    keywords and options may have a leading '-'. ");
fprintf(stderr,"Quick-edit a parameter\n");
fprintf(stderr,"    file by e.g. %s <file> %s <keyword> <value> %s <file>.\n",
					LOADPARMOPT, FLASHPARMOPT, SAVEPARMOPT);
fprintf(stderr,"\n");
}
/*----------------------------------------------------------------------*/
/*  SetVarFromKeytab							*/
/*									*/
/*		Besetzt die in der Keytab spezifierten Variablen mit	*/
/*		den aktuelle Keytab-Werten und ueberprueft ihre		*/
/*		Wertebereiche.						*/
/*		Bei Fehler Rueckgabestatus =! 0 und Text in *ERRMSGP.	*/
/*----------------------------------------------------------------------*/
static int SetVarFromKeytab(ktp, errmsgp)
KEY_TAB *ktp;
char **errmsgp;
{
	int i,count;
	double x;

	i = -1;
	while (ktp[++i].kwd != 0L) 
	{
	  /* Keytab-Variablen mit Keytab-Defaults besetzen */
	  if (StringCompare1(ktp[i].kfmt, STRING)) {
	    if (strlen(ktp[i].kval) > MAXLINE - 1) {
	      *errmsgp = 
	        catmsg(ktp[i].kwd, "string length >= MAXLINE");
	      return(-1);
	    }
	    strcpy(ktp[i].kvar,ktp[i].kval);	   
          } else {
	    if (sscanf(ktp[i].kval, ktp[i].kfmt, ktp[i].kvar) != 1) {
	      *errmsgp = catmsg(ktp[i].kwd, ktp[i].kval);
	      *errmsgp = catmsg(*errmsgp,"wrong argument type");
	      return(-1);
	    }
	  }
        }

	/* Wertebereichsueberpruefung	*/
	if(CheckInputRange(ktp, errmsgp))
		return(-1);
	return(0);
}
/*----------------------------------------------------------------------*/
/*  catmsg	fuegt die beiden Meldungen S1 und S2 mit einem ": "	*/
/*		in der Mitte zusammen, gibt Zeiger auf Ergebnisstring	*/
/*		zurueck. Puffer fuer den Ergebnisstring werden intern	*/
/*		bereitgestellt; durch alternierenden Wechsel zwischen	*/
/*		zwei Puffern ist mehrfach verschachtelter Aufruf	*/
/*		moeglich.						*/
/*----------------------------------------------------------------------*/
char *catmsg(s1,s2)
char *s1,*s2;
{
	static int toggle = 0;
	static char buffer1[MAXLINE], buffer2[MAXLINE];
	int ctr;
	char *p, *buffer;

	if (toggle == 0) {		/* bei jedem Aufruf Zielpuffer	*/
		toggle = 1;		/* wechseln			*/
		p = buffer = buffer1;
	} else {
		toggle = 0;
		p = buffer = buffer2;
	}				/* Zaehler gegen Zielpuffer-	*/
	ctr = 0;			/* Ueberlauf			*/
	while (ctr++ < MAXLINE-5)	/* s1 in Zielpuffer uebertragen	*/
		if ((*p++ = *s1++) == 0)
			break;
	--p;
	*p++ = ':'; *p++ = ' ';		/* ": " zwischen s1 und s2	*/

	while (ctr++ < MAXLINE-1)	/* s2 in Zielpuffer uebertragen	*/
		if ((*p++ = *s2++) == 0)
			break;		/* falls Abbruch wg. Zielpuffer	*/
	*p = 0;				/* fast voll, null anfuegen	*/

	return(buffer);			/* Zielpufferadr. zurueck	*/
}
/*----------------------------------------------------------------------*/
/*  errorhandler	gibt Fehlermeldung ERRMSG auf stderr aus.	*/
/*----------------------------------------------------------------------*/
errorhandler(errmsg)
char *errmsg;
{
	fprintf(stderr, "%s: ", g_myname);
	fprintf(stderr, "%s\n", errmsg);
}
/*----------------------------------------------------------------------*/
/*  CmdlineToKeytab							*/
/*									*/
/*		Ueberprueft Kommandozeilen-Syntax, uebergeben im	*/
/*		Zeigerfeld ARGV der Laenge ARGC, auf korrekten ge-	*/
/*		brauch von Keywords und Options. Die Indizes der Op-	*/
/*		tions SHOWPARMOPT,FLASHPARMOPT,SAVEPARMOPT,		*/
/*		INTNDFTSOPT im ARGV-Zeigerfeld werden per SHOWP,FLASHP	*/
/*		SAVEP bzw. DFTSP zurueckgegeben, wenn sie gefunden	*/
/*		wurden, ansonsten wird kleiner 0 zugewiesen. Maximal 	*/
/*		MAXLOAD LOADPARMOPTS sind moeglich, ihre ARGV-Indizes 	*/
/*		werden im Array LOADP[] gespeichert, die Anzal in *NLP. */
/*		Argumente nach denjenigen Keywords, die in Keytab KTP	*/
/*		eingetragen sind, werden den Keyword-Werten der Keytab	*/
/*		zugewiesen. Keytab-Variablen werden auf Keytab-Werte	*/
/*		gesetzt und der Wertebereich ueberprueft.		*/
/*		Bei Fehler Rueckgabestatus =! 0 und Text in *ERRMSGP.	*/
/*----------------------------------------------------------------------*/
static int CmdlineToKeytab(argc, argv, ktp,
	showp, flashp, loadp, nlp, savep, dftsp, errmsgp)
int argc, *showp, *flashp, *loadp, *nlp, *savep, *dftsp;
char *argv[], **errmsgp;
KEY_TAB *ktp;
{
	int i, k;
	char *help;

	*showp = *savep = *dftsp = *flashp = -1;
	*nlp = 0;			/* Anzahl gef. LOADPARM-Opt.	*/
	i = 0;				/* naechstes Kmdz.-Argument..	*/
	while (i < argc) {
					/* Ueberpreufung auf Options	*/
					/* ohne Keytab-Bezug,		*/
					/* die keine Argumente haben	*/
	    if (StringCompare2(*(argv+i), SHOWPARMOPT)) {
		*showp = i;
		i += 1;
	    } else if (StringCompare2(*(argv+i), FLASHPARMOPT)) {
		*flashp = i;
		i += 1;
	    } else if (StringCompare2(*(argv+i), INTNDFTSOPT)) {
		*dftsp = i;
		i += 1;
	    } else if (StringCompare2(*(argv+i), NOLIMOPT)) {
		g_nolim = i;
		i += 1;
	    } else { 			/* Ueberpruefung auf Keytab-	*/
					/* Keywords bzw. Options ohne	*/
					/* Keytab-Bezug, jew. mit einem	*/
					/* Argument			*/
		if ((k = FindKtEntry(ktp, *(argv+i))) >= 0) {
		    help = ktp[k].kwd;
		} else if (StringCompare2(*(argv+i), LOADPARMOPT)) {
		    if (*nlp < MAXLOAD)
		      loadp[(*nlp)++]=i;/* Position in argv zurueckgeb.	*/
		    help = LOADPARMOPT;	/* help fuer Argumentcheck	*/
		} else if (StringCompare2(*(argv+i), SAVEPARMOPT)) {
		    *savep = i;
		    help = SAVEPARMOPT;
		} else {
		    *errmsgp = catmsg(*(argv+i),"unknown keyword");
		    return(-1);
		}
		i += 2;
		if (i > argc) {		/* Argumentcheck		*/
		    *errmsgp = catmsg(help,"argument missing");
		    return(-1);
		}			/* Zeiger Keytab-Wert auf	*/
		if (k >= 0)		/* Argument umbiegen		*/
		    ktp[k].kval = *(argv+i-1);
			
	    }				/* Keytab-Variablen Setzen	*/
	}				/* und Bereichsueberpruefung	*/
	return(SetVarFromKeytab(ktp, errmsgp));
}
/*----------------------------------------------------------------------*/
/*  ParmFileToKeyTab							*/
/*									*/
/*		Oeffnet und liest Parameterdatei FNAME. Weist die	*/
/*		hinter Keywords gefundenen Argumente der Keytab	KTP zu.	*/
/*		Rueckgabestatus 0 fuer OK, <0 fuer allgemeiner Fehler	*/
/*		und >0 fuer Parameterdatei nicht vorhanden. Fehlertext	*/
/*		in *ERRMSGP.						*/
/*----------------------------------------------------------------------*/
static int ParmFileToKeytab(fname, ktp, errmsgp)
char *fname, **errmsgp;
KEY_TAB *ktp;
{
	FILE *fp;

	if ((fp = fopen(fname, "r")) == NULL) {
		*errmsgp = catmsg(fname,"cannot open");
		return(1);
	}
	if (StreamToKeytab(fp, ktp, errmsgp)) {
		fclose(fp);
		*errmsgp = catmsg(fname, *errmsgp);
		return(-1);
	}
	if (fclose(fp)) {
		*errmsgp = catmsg(fname,"cannot close");
		return(-1);
	}
	return(0);
}
/*----------------------------------------------------------------------*/
/*   KeytabToParmFile							*/
/*									*/
/*		Schreibt Keytab KTP in durch ParmFileToKeytab wieder-	*/
/*		einlesbarer Form in die	Datei FNAME.			*/
/*		Bei Fehler Rueckgabestatus =! 0 und Text in *ERRMSGP.	*/
/*----------------------------------------------------------------------*/
static int KeytabToParmFile(fname, ktp, errmsgp)
char *fname, **errmsgp;
KEY_TAB *ktp;
{
	FILE	*fp;

	if ((fp = fopen(fname, "w")) == NULL) {
		*errmsgp = catmsg(fname,"cannot open");
		return(-1);
	}
	if (KeytabToStream(fp,ktp,errmsgp)) {
		fclose(fp);
		*errmsgp = catmsg(fname,*errmsgp);
		return(-1);
	}
	
	if (fclose(fp)) {
		*errmsgp = catmsg(fname,"cannot close");
		return(-1);
	}
	return(0);
}
/*----------------------------------------------------------------------*/
/*  LongInfoParm							*/
/*									*/
/*		gibt Parameter auf Bildschirm aus,			*/
/*		ausfuehrliche Darstellung.				*/
/*----------------------------------------------------------------------*/
static LongInfoParm(ktp)
KEY_TAB *ktp;
{
	char *errmsg;
					/* Fehler nichtssagend		*/
	KeytabToStream(stdout, ktp, &errmsg);
}
/*----------------------------------------------------------------------*/
/*  InfoParm								*/
/*									*/
/*		gibt Parameter auf Bildschirm aus, Kurzinfo.		*/
/*----------------------------------------------------------------------*/
static InfoParm(ktp)
KEY_TAB *ktp;
{
	int i, j, l;

	i = -1;
	while (ktp[++i].kwd != 0L) {	/* ueber alle Keytab-Eintraege	*/
	    printf("%-15s ", ktp[i].kwd);
					/* falls Keytab-Format STRING..	*/
	    if (StringCompare1(ktp[i].kfmt, STRING))
	    {
                /* Einzelstrings ohne Anfuehrungszeichen,   */
		/* durch Leerzeichen getrennte Teilstrings  */ 
		/* in Anfuehrungszeichen                    */
	        l = strlen(ktp[i].kvar);
	        if (!strstr(ktp[i].kvar, " ")) {
		     printf("%s ", ktp[i].kvar);
		     if (l > 19)
		       printf("\n%36.36s", "");
		     else
		       for (j = l; j < 19; j++)
		         printf(" ");
	        } else {
		     printf("\"%s\" ", ktp[i].kvar);
		     if (l > 17)
		       printf("\n%36.36s", "");
		     else
		       for (j = l; j < 17; j++)
		         printf(" ");
                }
            }
	    else if (StringCompare1(ktp[i].kfmt, DOUBLE)) {
	      if (*(double *)ktp[i].kvar >= 0)
		printf("%-10.4le          ", *(double *)ktp[i].kvar);
	      else
		printf("%-+10.4le         ", *(double *)ktp[i].kvar);
	    } else {
	      printf("%-12d        ", *(int *)ktp[i].kvar);
	    }
	    printf("# %s\n", ktp[i].ktxt);
	}
}
/*----------------------------------------------------------------------*/
/*  CheckInputRange							*/
/*									*/
/*		Kontrolliert die Zulaessigkeit der Keytab-Werte in KTP	*/
/*		anhand der ebenfalls dort vorhandenen Wertebereichs-	*/
/*		angaben.						*/
/*		Bei Fehler Rueckgabestatus =! 0 und Text in *ERRMSGP.	*/
/*----------------------------------------------------------------------*/
static int CheckInputRange(ktp, errmsgp)
KEY_TAB *ktp;
char **errmsgp;
{
	int i, status, ui, oi, vi;
	double ud, od, vd;
	char *p, *q, *e, *d;
	static char string[MAXLINE+1];

	if (g_nolim >= 0)
	    return(0);			/* keine Zulaessigkeitspruefung	*/
	i = -1;				/* Ueber alle Keytab-Eintraege	*/
	while (ktp[++i].kwd != 0L) {	/* falls Bereich spezifiziert	*/
	    if (!StringCompare1(ktp[i].krge, EMPTY)) {
	   	p = ktp[i].krge;	/* Wertebereichsstringzeiger/	*/
					/* Zulassungsmusterzeiger an    */
					/* ersten WBS/ZM                */
					/* falls Format STRING		*/
		if (StringCompare1(ktp[i].kfmt, STRING)) {
		    d = " ";		/* Unterteiler Zulassungsmuster */
		    if (strstr(p, "*")) /* falls alternativer Untert.	*/
			d = "*";        /* vorkommt, diesen verwenden   */
		                        /* Zeiger auf Ende aller ZM     */
		    e = ktp[i].krge + strlen(ktp[i].krge);
		    for (;;) {
			if (p >= e) {
			    *errmsgp = catmsg(ktp[i].kwd, ktp[i].kval);
			    *errmsgp =
			        catmsg(*errmsgp, "value not allowed");
			    return (-1);
			}
			q = strstr(p, d);
			if (!q)
			    q = e;
			strncpy(string, p, q - p);
			string[(int)(q - p)] = 0;
			if (StringCompare1(ktp[i].kval, string))
			    break;
			p = q + 1;
		    }			/* Falls Format DOUBLE		*/
		} else if (StringCompare1(ktp[i].kfmt, DOUBLE)) {
					/* Variablenwert holen und	*/
					/* in String zurueckwandeln	*/
		    vd = *(double*)ktp[i].kvar;
		    sprintf(string, "%8le", vd);
					/* Wertebereich aus Wertebe-	*/
					/* reichsstring holen, es	*/
					/* muessen Anfangs und		*/
					/* Endwert angegeben sein	*/
		    if (sscanf(p, "%lf%lf", &ud, &od) != 2 || ud > od) {
			*errmsgp = catmsg(ktp[i].kwd, p); 
			*errmsgp = catmsg(*errmsgp,
			    "wrong range specification");
			return(-1);
					/* Wertebereichsueberpruefung	*/
		    } else if (vd > od || vd < ud) {
			*errmsgp = catmsg(ktp[i].kwd, string);
			*errmsgp = catmsg(*errmsgp, 
			    "out of range");
			return(-1);
		    }
					/* Format INT analog DOUBLE	*/
		} else if (StringCompare1(ktp[i].kfmt, INT)) {
		    vi = *(int*)ktp[i].kvar;
		    sprintf(string, "%d", vi);
		    if (sscanf(p, "%d%d", &ui, &oi) != 2 || ui > oi) {
			*errmsgp = catmsg(ktp[i].kwd, p); 
			*errmsgp = catmsg(*errmsgp,
			    "wrong range specification");
			return(-1);
		    }
		    if (vi > oi || vi < ui) {
			*errmsgp = catmsg(ktp[i].kwd, string);
			*errmsgp = catmsg(*errmsgp, 
			    "out of range");
			return(-1);
		    }
	        } else {		/* interne Fehlerabsicherung	*/
		    *errmsgp = catmsg(ktp[i].kwd, ktp[i].kfmt); 
		    *errmsgp = catmsg( *errmsgp,
			"undefined format");
		    return(-1);
		}
	    }
	}
	return(0);
}
/*----------------------------------------------------------------------*/
/*  FindKtEntry	gibt den Tabellenindex des Keywords KWD in der		*/
/*		Tabelle KTP zurueck. Rueckgabe kleiner null, falls	*/
/*		KWD nicht in KTP enthalten.				*/
/*									*/
/*  BEACHTE:	Gross- und Kleinschreibung des Keywords wird nicht	*/
/*		beachtet und es darf ein '-' vorangestellt sein.	*/
/*----------------------------------------------------------------------*/
static int FindKtEntry(ktp, kwd)
KEY_TAB *ktp;
char *kwd;
{
	int i;

	i = -1;
	while (ktp[++i].kwd != 0L)
	    if (StringCompare2(kwd, ktp[i].kwd))
		return(i);
	return(-1);
}
/*----------------------------------------------------------------------*/
/*  StringCompare1							*/
/*									*/
/*		Vergleicht Strings miteinander. Bei Nichtuebereinstim-	*/
/*		mumg wird null zurueckgegeben.				*/
/*									*/
/*  BEACHTE:	Gross- und Kleinschreibung der Strings wird nicht	*/
/*		beachtet.						*/
/*----------------------------------------------------------------------*/
static int StringCompare1(s1, s2)
register char *s1, *s2;
{
	register int i;

	i = 0;
					/* String-Compare, ohne Gross-	*/
	while (s1[i] != 0) {		/* und Kleinschreibung		*/
		if ((s1[i] | 0x20) != (s2[i] | 0x20)) 
			break;
		i++;
	}			
    	if (s1[i] == 0 && s2[i] == 0)
		return(1);		/* Uebereinstimmung		*/
	return(0);			/* Misserfolg			*/
}
/*----------------------------------------------------------------------*/
/*  StringCompare2							*/
/*									*/
/*		Vergleicht Strings miteinander. Bei Nichtuebereinstim-	*/
/*		mumg wird null zurueckgegeben.				*/
/*									*/
/*  BEACHTE:	Gross- und Kleinschreibung der Strings wird nicht be-	*/
/*		achtet, einem String darf ein '-' vorangestellt sein.	*/
/*----------------------------------------------------------------------*/
static int StringCompare2(s1, s2)
register char *s1, *s2;
{
	register int i,j;

	i = 0;
	if (s1[0] == '-')		/* ignoriere "Options '-'"	*/
		s1++;
	if (s2[0] == '-')
		s2++;
	return(StringCompare1(s1, s2));
}
/*----------------------------------------------------------------------*/
/*  StreamToKeytab							*/
/*									*/
/*		Liest STREAM und sucht dort Keywords in Uebereinstim-	*/
/*		mung mit der Keytab KTP. Uebertragt die den gefundenen	*/
/*		Keywords nachfolgenden Argumente in einen Puffer und	*/
/*		biegt den entsprechenden Keytab-Wert-Zeiger um auf den	*/
/*		Pufferplatz. Die Keytab-Variablen werden auf die	*/
/*		Keytab-Werte gesetzt und der Wertebereich ueberprueft.	*/
/*		Bei Fehler Rueckgabestatus =! 0 und Text in *ERRMSGP.	*/
/*----------------------------------------------------------------------*/
static int StreamToKeytab(stream, ktp, errmsgp)
FILE *stream;
char **errmsgp;
KEY_TAB *ktp;
{
	static char kwd[MAXLINE+1], karg[MAXLINE+1];
	char  *p;
	int i;
					/* Kwd/Karg-Paare bis EOF	*/
	while (!NextKwdKargStream(stream, kwd, karg, MAXLINE)) {
					/* wenn Keyword bekannt:	*/
	    if ((i = FindKtEntry(ktp, kwd)) >= 0) {
		if (*karg == 0 && !StringCompare1(ktp[i].kfmt, STRING)) {
		    *errmsgp = catmsg(ktp[i].kwd,"argument missing");
		    return(-1);
		}			/* Speicher Keyword-Argument	*/
		p = (char *)malloc(strlen(karg)+1);
		if (!p) {
			*errmsgp = "keyword manager out of memory";
			return(-1);
		}
		strcpy(p, karg);	/* Keyword-Argument uebertragen	*/
		  			/* Zeiger Keyword-Arg umbiegen	*/
		ktp[i].kval = p;	/* auf Argumentplatz		*/
	    }
					/* Keytab-Variablen Setzen	*/
	}				/* und Bereichsueberpruefung	*/
	return(SetVarFromKeytab(ktp, errmsgp));
}
/*----------------------------------------------------------------------*/
/*  KeytabToStream							*/
/*									*/
/*		Ausgabe Keytab KTP in wiedereinlesbarer Form auf STREAM.*/
/*		Bei Fehler Rueckgabestatus =! 0 und Text in *ERRMSGP.	*/
/*----------------------------------------------------------------------*/
static int KeytabToStream(stream, ktp, errmsgp)
FILE *stream;
KEY_TAB *ktp;
char **errmsgp;
{
	int i, ui, oi, flag;
	double ud, od;
	static char string[MAXLINE+1];
	char *p, *q, *d, *e, c;

	i = -1;
	while (ktp[++i].kwd != 0L) {	/* ueber alle Keytab-Eintraege	*/
					/* Keytab-Text ausgeben		*/
	    fprintf(stream,"# %s\n", ktp[i].ktxt);
					/* falls Keytab-Format STRING	*/
	    if (StringCompare1(ktp[i].kfmt, STRING)) {

		if (strstr(ktp[i].kvar,","))
		    flag = 1;		/* Quotierung vormerken		*/

					/* wenn Wertebereich/Zulassungs-*/
					/* muster (ZM) spezifiziert	*/
		if (!StringCompare1(ktp[i].krge, EMPTY)) {
		    c = '{';		/* ZM-Ausdrucktrenner, Beginn  	*/
		    fprintf(stream,"# ");
		    p = ktp[i].krge;
		    flag = 0;		/* Quotierungsflag "aus"	*/
		    d = " ";		/* ZM-Trenner			*/
		    if (strstr(p,"*")){ /* falls alternativer Untert.	*/
			d = "*";        /* vorkommt, diesen verwenden   */
			flag = 1;
		    }			/* Wenn Komma oder Leerzeichen  */
		    if (strstr(p,","))  /* im ZM, Quotierung vormerken  */
		        flag = 1;
		                        /* Zeiger auf Ende aller ZM     */
		    e = ktp[i].krge + strlen(ktp[i].krge);
		    for (;;) {
			if (p >= e)
			    break;
			q = strstr(p, d);
			if (!q)
			    q = e;
			strncpy(string, p, q - p);
			string[(int)(q - p)] = 0;
			if (flag)
			    fprintf(stream,"%c\"%s\"", c, string);
		        else
			    fprintf(stream,"%c%s", c, string);
			c = ',';
			p = q + 1;
		    }
	    	    fprintf(stream,"}\n#\n");
		} else			/* kein ZM-String vorhanden	*/
	    	    fprintf(stream,"#\n#\n");
		if (strstr(ktp[i].kvar," "))
		    			/* Wenn im Varaiablenwert Leer-	*/
					/* zeichen Komma vorkommen:	*/                
		    fprintf(stream,"%-15.15s \"%s\"",
			    ktp[i].kwd,ktp[i].kvar);
		else
		    fprintf(stream,"%-15.15s %s",
			    ktp[i].kwd, ktp[i].kvar);
					/* falls Keytab-Format DOUBLE	*/
	    } else if (StringCompare1(ktp[i].kfmt, DOUBLE)) {
					/* falls Wertebereich (WB) spez.*/
		if (!StringCompare1(ktp[i].krge, EMPTY)) {
					/* WB-Start und -Ende Ausgeben	*/
		    sscanf(ktp[i].krge, "%lf%lf", &ud, &od);
	    	    fprintf(stream,"# [%le...%le]\n#\n",ud,od);
		} else			/* falls kein WB spezifiziert	*/
	    	    fprintf(stream,"#\n#\n");
		fprintf(stream,"%-15.15s %le",
		     ktp[i].kwd, *(double *)ktp[i].kvar);
	    } else {			/* Format INT analog DOUBLE...	*/
		if (!StringCompare1(ktp[i].krge, EMPTY)) {
		    sscanf(ktp[i].krge, "%d%d", &ui, &oi);
	    	    fprintf(stream,"# [%d...%d]\n#\n", ui, oi);
		} else
	    	    fprintf(stream,"#\n#\n");
		fprintf(stream,"%-15.15s %d",
		     ktp[i].kwd, *(int *)ktp[i].kvar);
	    }
	    fprintf(stream,"\n\n");
	    if (feof(stream) || ferror(stream)) {
		*errmsgp = "cannot write";
		return(-1);
	    }
	}
	return(0);
}
/*----------------------------------------------------------------------*/
/* NextKwdKargStream	holt aus Stream naechstes Keyword/Argument-	*/
/*			Paar, indem naechste Nicht-Weiss-Zeichenfolge	*/
/*			als Keyword interpretiert wird. Die zweite	*/
/*			Nicht-Weiss-Zeichenfolge auf gleicher Zeile	*/
/*			wird als Argument interpretiert. Mit '#'	*/
/*			beginnende Zeilenreste werden ignoriert.	*/
/*			Rueckgabe != 0 bei EOF / kein Keyword gefunden.	*/
/*			Sonst karg Nullstring, wenn kein Argument gef.	*/
/*									*/
/*	stream		Stream-Zeiger					*/
/*	kwd, karg	Zeiger fuer Keyword- und Argumentstring		*/
/*	maxlen		maximale Laenge (>=1) von kwd, karg incl. '\0'	*/
/*----------------------------------------------------------------------*/
static int NextKwdKargStream(stream, kwd, karg, maxlen)
FILE *stream;
char *kwd, *karg;
int maxlen;
{
	int i = 0, j = 0, c;
 	
	c = fgetc(stream);
	while (c != EOF) {		/* WIEDERHOLE BIS EOF		*/
					/* Absaugen bis			*/
					/* Nicht-Weiss-Zeichen oder EOF */
	  while (c==' '||c=='\t'||c=='\n'||c=='\r')
	    c = fgetc(stream);
	  if (c == '#') {		/* Kommentar: Zeilenrest weg...	*/
	    do {
	      c = fgetc(stream);
	    } while (c != '\n' && c != EOF);
	    while (c == '\n' || c== '\r')
	      c = fgetc(stream);	/* ...wenn nicht EOF, von vorne	*/
	  } else {			/* kwd einlesen (maxlen!) bis	*/
					/* Weiss-Zeichen oder EOF	*/
	    while (c!=' '&&c!='\t'&&c!='\n'&&c!='\r'&&c!='#'&&c!=EOF) {
	      if (i < maxlen)
		kwd[i++] = c;
	      c = fgetc(stream);
	    }
	    while (c==' '||c=='\t'||c=='\n'||c=='\r')
	      c = fgetc(stream);	/* weg bis Nicht-Weiss-Zeichen	*/
	    if (c == '#') {		/* Kommentar: Zeilenrest weg... */
	      do {
		c = fgetc(stream);
	      } while (c != '\n' && c != EOF);
	      while (c == '\n' || c == '\r')
		c = fgetc(stream);	/* ... karg muss leerbleiben	*/
	    } else {			/* Kein Kommentar:		*/
	      if (c == '"') {		/* Erstes Nicht-Weiss-Zeichen	*/
		c = fgetc(stream);	/* ist `"' ...			*/
		while (c!='"'&&c!='\n'&&c!='\r'&&c!='#'&&c!=EOF) {
		  if (j < maxlen)
		    karg[j++] = c; 
		  c = fgetc(stream);
		}			/* ... karg einlesen bis `"',	*/
					/* Zeilenende oder EOF.		*/
	      } else {			/* Erstes Nicht-Weiss-Zeichen	*/
					/* ist kein `"' ...		*/
		while (c!=' '&&c!='\t'&&c!='\n'&&c!='\r'&&c!='#'&&c!=EOF){
		  if (j < maxlen)
		    karg[j++] = c; 
		  c = fgetc(stream);
		}			/* ... karg einlesen bis	*/
	      }				/* Weiss-Zeichen oder EOF	*/
	      while (c != '\n' && c != '\r' && c != EOF)
		c = fgetc(stream);	/* Zeilenrest absaugen		*/
	    }
	    karg[j] = 0;
	    kwd[i] = 0;
	    return(0);
	  }
	}
	return(-1);
}
/*----------------------------------------------------------------------*/
/*  convspchar	wandelt in Strings Escape-Sequenzen in 'C'-Style um	*/
/*		in die entsprechenden Binaerwerte. Interner Puffer.	*/
/*----------------------------------------------------------------------*/
char *convspchar(p)
char *p;
{
	static char out[MAXLINE];
	char *q, *qmax;

	q = out;
	qmax = q + MAXLINE;

	while (*p != 0 && q != qmax) {
	    if (*p == '\\' ) {
		switch (*++p) {
		case '\\': *q++ = '\\'; p++; break;
		case 'n' : *q++ = '\n'; p++; break;
		case 't' : *q++ = '\t'; p++; break;
		case 'r' : *q++ = '\r'; p++; break;
		default:
		    if ((*p     >= '0' && *p     <= '7') &&
		        (*(p+1) >= '0' && *(p+1) <= '7') &&
		        (*(p+2) >= '0' && *(p+2) <= '7')) {
		      *q++ = 
			(*p -'0')*64 + (*(p+1)-'0')*8 + *(p+2)-'0';
		      p += 3;
		    }
		}
	    } else
		*q++ = *p++;	
	}
	*q = 0;
	return(out);
}
/*******Ende*************************************************************/
