/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/*************************************************************************
 *
 *   Name    : numio.c
 *   Version : 1999/08/03 1.2
 *   Comment : numeric data I/O
 *   Author  : Markus Mummert
 *   History :
 *   1999/07/26 begin coding
 *   1999/08/03 1.2: multiple include protection in numio.h
 *
 */

#include <stdio.h>
#include <math.h>
#include "portab.h"
#include "numio.h"

/*
 *  numread
 *
 *  Read <count> numbers from <stream> according to <fmt>, convert into
 *  double, and store into <buff>. Returns number of numbers actually
 *  read. Formats specified by <fmt> are discerned by their first letter
 *  (may be capital or small), defined by constants NUMIO_xxx_FMT_ID,
 *  xxx = ASCII, DOUBLE, FLOAT, BYTE, UBYTE, WORD, LONG.
 *
 *  C type identifiers for the formats xxx = BYTE, UBYTE, WORD, LONG are
 *  defined by constants NUMIO_xxx. Numbers read in these Formats 
 *  will be divided by constants NUMIO_xSCALE, x = b, u, w, l before
 *  they are stored into <buff>.
 *
 *  About conversions see comment in .h-file.
 *
 */
int numread(stream, fmt, buff, count)
FILE *stream;
char *fmt;
double *buff;
int count;
{
	float *f;
	NUMIO_UBYTE *u;
	NUMIO_BYTE *b;
	NUMIO_WORD *w;
	NUMIO_LONG *l;
	int i, k;
	double ilscale, iwscale, ibscale, iuscale;

        switch(fmt[0] | 0x20) {
	    case NUMIO_ASCII_FMT_ID: 
		for (i = 0; i < count; i++)
	            if (fscanf(stream, "%lf", buff+i) <= 0)
		        break;
	        break;
	    case NUMIO_FLOAT_FMT_ID:
		f = (float*)buff;
	       	i = FREAD(f, sizeof(float), count, stream);
		for (k = i - 1; k >= 0; k--)
		    buff[k] = (double)(f[k]);
	        break;
	    case NUMIO_DOUBLE_FMT_ID:
	        i = FREAD(buff, sizeof(double), count, stream);
	        break;
	    case NUMIO_BYTE_FMT_ID:
		ibscale = 1.0 / NUMIO_BSCALE;
		b = (NUMIO_BYTE*)buff;
	       	i = FREAD(b, sizeof(NUMIO_BYTE), count, stream);
		for (k = i - 1; k >= 0; k--)
		    buff[k] = b[k] * ibscale;
	        break;
	    case NUMIO_UBYTE_FMT_ID:
		iuscale = 1.0 / NUMIO_USCALE;
		u = (NUMIO_UBYTE*)buff;
	       	i = FREAD(u, sizeof(NUMIO_UBYTE), count, stream);
		for (k = i - 1; k >= 0; k--)
		    buff[k] = u[k] * iuscale;
	        break;
	    case NUMIO_WORD_FMT_ID:
		iwscale = 1.0 / NUMIO_WSCALE;
		w = (NUMIO_WORD*)buff;
	       	i = FREAD(w, sizeof(NUMIO_WORD), count, stream);
		for (k = i - 1; k >= 0; k--)
		    buff[k] = w[k] * iwscale;
	        break;
	    case NUMIO_LONG_FMT_ID:
		ilscale = 1.0 / NUMIO_LSCALE;
		l = (NUMIO_LONG*)buff;
	       	i = FREAD(l, sizeof(NUMIO_LONG), count, stream);
		for (k = i - 1; k >= 0; k--)
		    buff[k] = l[k] * ilscale;
	        break;
	    default:
		fprintf(stderr, 
		  "UNEXPECTED: numread(): undefined format `%c'\n",
		  fmt[0]
		);
		exit(-1);
		break;
        }
	return(i);
}
/*
 *  numwrite
 *
 *  write <count> numbers of type double from <buff> to <stream> in a
 *  format specified by <fmt>. Returns number of numbers actually
 *  written, possibly preceeded by a minus sign if overflow occured.
 *  Formats are discerned by their first letter as with numread().
 *  Exept for ASCII- and DOUBLE-format, the contents of <buff> will
 *  be lost upon return, due to format converion.
 *
 *  Complementary to numread() numbers for fixed-point binary
 *  formats are multiplied by constants NUMIO_xSCALE, x = b, u, w, l
 *  before they are converted. Also, for these types a type-range
 *  overflow-condition is checked. Written values will be clipped (!)
 *  to the largest or smallest number fitting into the range.
 *
 *  For ASCII-format, <fields> is the number of fields written
 *  in one line, each specified by fprintf()-format <afieldfmt>.
 *  More than one fields in a line are separated by <afieldsep>,
 *  The line is terminated by NUMIO_LINESEP.
 *
 *  About conversions see comment in .h-file.
 *
 */
int numwrite(stream, fmt, buff, count, afields, afieldfmt, afieldsep)
FILE *stream;
char *fmt, *afieldfmt, *afieldsep;
double *buff;
int count, afields;
{
	int i, k;
	float *f;
	NUMIO_BYTE cb, *b;
	NUMIO_UBYTE cu, *u;
	NUMIO_WORD cw, *w;
	NUMIO_LONG cl, *l;
	char *alinesep = "\n";
	int chunk, n, m, ovfl;
	double cd, lscale, wscale, bscale, uscale;

	i = 0;
	ovfl = 0;
	switch(fmt[0] | 0x20) {
	    case NUMIO_ASCII_FMT_ID: 
		chunk = afields;
		for (m = 0; m < count && !ferror(stream); m += chunk) {
		    n = (count - m < chunk) ? count - m : chunk;
		    for (k = 0; k < n; k++) {
			if (k)
			    fprintf(stream, afieldsep);
		        fprintf(stream, afieldfmt, buff[k+m]);
			if (ferror(stream))
			    break;
			i++;
		    }
		    fprintf(stream, alinesep);
		}
	        break;
	    case NUMIO_FLOAT_FMT_ID:
		f = (float*)buff;
		chunk = NUMIO_SIZEOFWCHUNK/sizeof(float);
		for (m = 0; m < count; m += chunk) {
		    n = (count - m < chunk) ? count - m : chunk;
		    for (k = 0; k < n; k++) 
		        f[k+m] = (float)(buff[k+m]);
	       	    i += FWRITE(f+m, sizeof(float), n, stream);
		}
	        break;
	    case NUMIO_DOUBLE_FMT_ID:
		chunk = NUMIO_SIZEOFWCHUNK/sizeof(double);
		for (m = 0; m < count; m += chunk) {
		    n = (count - m < chunk) ? count - m : chunk;
	       	    i += FWRITE(buff+m, sizeof(double), n, stream);
		}
	        break;
	    case NUMIO_BYTE_FMT_ID:
		bscale = NUMIO_BSCALE;
		b = (NUMIO_BYTE*)buff;
		chunk = NUMIO_SIZEOFWCHUNK/sizeof(NUMIO_BYTE);
		for (m = 0; m < count; m += chunk) {
		    n = (count - m < chunk) ? count - m : chunk;
		    for (k = 0; k < n; k++) {
		        b[k+m] = (cb = (NUMIO_BYTE)
				  (floor(cd = buff[k+m] * bscale)));
			if (fabs(cb - cd) >= 1.0) {
			    ovfl = 1;
			    if (cd > 0)
				b[k+m] = NUMIO_BMAX;
			    else
				b[k+m] = (NUMIO_BYTE)~NUMIO_BMAX;
			}
		    } 
	       	    i += FWRITE(b+m, sizeof(NUMIO_BYTE), n, stream);
		}
	        break;
	    case NUMIO_UBYTE_FMT_ID:
		uscale = NUMIO_USCALE;
		u = (NUMIO_UBYTE*)buff;
		chunk = NUMIO_SIZEOFWCHUNK/sizeof(NUMIO_UBYTE);
		for (m = 0; m < count; m += chunk) {
		    n = (count - m < chunk) ? count - m : chunk;
		    for (k = 0; k < n; k++) {
		        u[k+m] = (cu = (NUMIO_UBYTE)
				  (floor(cd = buff[k+m] * uscale)));
			if (fabs(cu - cd) >= 1.0) {
			    ovfl = 1;
			    if (cd > 0)
				u[k+m] = NUMIO_UMAX;
			    else
				u[k+m] = (NUMIO_UBYTE)~NUMIO_UMAX;
			}
		    }
	       	    i += FWRITE(u+m, sizeof(NUMIO_UBYTE), n, stream);
		}
		break;
	    case NUMIO_WORD_FMT_ID:
		wscale = NUMIO_WSCALE;
		w = (NUMIO_WORD*)buff;
		chunk = NUMIO_SIZEOFWCHUNK/sizeof(NUMIO_WORD);
		for (m = 0; m < count; m += chunk) {
		    n = (count - m < chunk) ? count - m : chunk;
		    for (k = 0; k < n; k++) {
		        w[k+m] = (cw = (NUMIO_WORD)
				  (cd = floor(buff[k+m] * wscale)));
			if (fabs(cw - cd) >= 1.0) {
			    ovfl = 1;
			    if (cd > 0)
				w[k+m] = NUMIO_WMAX;
			    else
				w[k+m] = (NUMIO_WORD)~NUMIO_WMAX;
			}
		    }
	       	    i += FWRITE(w+m, sizeof(NUMIO_WORD), n, stream);
		}
	        break;
	    case NUMIO_LONG_FMT_ID:
		lscale = NUMIO_LSCALE;
		l = (NUMIO_LONG*)buff;
		chunk = NUMIO_SIZEOFWCHUNK/sizeof(NUMIO_LONG);
		for (m = 0; m < count; m += chunk) {
		    n = (count - m < chunk) ? count - m : chunk;
		    for (k = 0; k < n; k++) {
		        l[k+m] = (cl = (NUMIO_LONG)
				  (cd = floor(buff[k+m] * lscale)));
			if (fabs(cl - cd) >= 1.0) {
			    ovfl = 1;
			    if (cd > 0)
				l[k+m] = NUMIO_LMAX;
			    else
				l[k+m] = (NUMIO_LONG)~NUMIO_LMAX;
			}
		    }
	       	    i += FWRITE(l+m, sizeof(NUMIO_LONG), n, stream);
		}
	        break;
	    default:
		fprintf(stderr, 
		  "UNEXPECTED: numwrite(): undefined format `%c'\n",
		  fmt[0]
		);
		exit(-1);
		break;
	}
	if (ovfl)
		return(-i);
	return(i);
}

/*
 * EOT
 *
 ************************************************************************/

