/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/*************************************************************************
 *
 *   Name    : numio.h
 *   Version : 1999/08/03 1.2
 *   Comment : numeric data I/O 
 *   Author  : Markus Mummert
 *   History : (see .c-file)
 *
 */
#ifndef NUMIO_INCLUDED	/* multiple include protection */
#define NUMIO_INCLUDED

/*
 *  Formats (capital or small letters): 
 *
 *      a(scii)   f(loat)  d(ouble)   b(yte)
 *	u(byte)   w(ord)   l(ong)
 *
 */
#define NUMIO_ASCII_FMT_ID	'a'
#define NUMIO_DOUBLE_FMT_ID	'd'
#define NUMIO_FLOAT_FMT_ID	'f'
#define NUMIO_BYTE_FMT_ID	'b'
#define NUMIO_UBYTE_FMT_ID	'u'
#define NUMIO_WORD_FMT_ID	'w'
#define NUMIO_LONG_FMT_ID	'l'

/*
 *  C Types associated with formats
 *  (taking defines from portab.h; float and double my not be changed)
 *
 */
#define NUMIO_BYTE		BYTE /* should be 8bit signed */
#define NUMIO_UBYTE		unsigned BYTE /* should be 8bit unsigned */
#define NUMIO_WORD		WORD /* should be 16bit signed */
#define NUMIO_LONG		LONG /* should be 32bit signed */

/*
 *  Scaling factors for integer formats. The idea is to have double 
 *  values in the range from -1.0 [or 0.0 if unsigned] to +1.0 to be 
 *  transformed into the full range of the integer formats. As a scaling
 *  factor the maximum-positive value of an integer format plus 1 is used
 *  (e.g. 1+(double)0x7FFF=2^15 for the WORD-format). When converting
 *  back to double the inverse scaling factor is taken. The result is:
 *  
 *  a) The limit -1.0 corresponds to the minimum-negative value (e.g.
 *     -2^15), but the limit +1.0 is just out of range. A floating point
 *     sine generator thus my not use 1.0 as full amplitude, but 
 *     1.0 minus a small NUMIO_DELTA !!! (1E-10 would be reasonable).
 *  b) Reading from an integer format results in a floating point range
 *     of -1.0 [or 0.0] to 1.0-1.0/S, with S as original scaling factor.
 *  c) Subsequend conversion back and forth will not change the signal.
 *
 *  Note that conversion to integer formats is obtained by a (type)(floor(.))
 *  construct, rather than just a cast (type)(.) which would produce (on
 *  some compilers?) a doubled quantize error around zero.
 *
 */
#define NUMIO_DELTA		1E-10

#define NUMIO_BMAX		0x7F
#define NUMIO_UMAX		0xFF
#define NUMIO_WMAX		0x7FFF
#define NUMIO_LMAX		0x7FFFFFFF

#define NUMIO_BSCALE		(1+(double)NUMIO_BMAX)
#define NUMIO_USCALE		(1+(double)NUMIO_UMAX)
#define NUMIO_WSCALE		(1+(double)NUMIO_WMAX)
#define NUMIO_LSCALE		(1+(double)NUMIO_LMAX)

#define NUMIO_LINESEP		"\n" /* line-separator for ASCII */
#define NUMIO_SIZEOFWCHUNK      512 /* sizeof chunks when calling FWRITE, */
                                    /* maybe not very useful .. */

extern int numread(/* stream, fmt, buff, count */);
extern int numwrite(/* stream, fmt, buff, count,
	 afields, afieldfmt, afieldsep */);

#endif /* NUMIO_INCLUDED */
/*
 * EOT
 *
 ************************************************************************/

