/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/*************************************************************************
 *
 *   Name    : window.c
 *   Version : 1999/08/20 1.15
 *   Comment : Data windows for FIR-design and signal analysis
 *   Author  : Markus Mummert
 *   History :
 *   1999/08/03 begin coding
 *
 */

#include <stdio.h>
#include <math.h>
#include "portab.h"
#include "filtflp.h"
#include "fft.h"
#include "window.h"

#define I0(x) I0_iter(x) /* I0_poly(x) */ /* select procedure */

/*
 *  Modified Bessel-function of zeroth order, iteration
 */
#define I0EPSILON 1e-20	 /* 1e-20: 100/30/20 iterations |x| < 100/10/5	*/
static double I0_iter(x) /* I0EPSILON hasn't too much impact on speed   */
double x;
{
	double i = 0.0, x2 = x/2.0, one = 1.0, frac = 1.0, term,
		sum = 1.0, eps=I0EPSILON;
	do {
		frac *= x2/(i+=one);
		term = frac*frac;
	} while(term >= eps*(sum+=term));
	return(sum);
}

/*
 *  Modified Bessel-function of zeroth order, polynominal fit,
 *  Abramowitz, Stegun, Handbook of Mathematical Functions,
 *  Dover Publications, Inc., New York, 1972.
 *  Within specified accuracy this may be faster than iteration
 */
double I0_poly(x)
double x;
{
	double t, tinv, tqua;

	t = x/3.75;
	tinv = 1.0/t;
	tqua = t*t;

	if (fabs(t) <= 1.0) {
	    return( /* +epsilon		fabs(epsilon) < 1.6E-7 */
		      (+1.0000000
		+tqua*(+3.5156229
		+tqua*(+3.0899424
		+tqua*(+1.2067492
		+tqua*(+0.2659732
		+tqua*(+0.0360768
		+tqua*(+0.0045813)))))))
	    );
	} else {
	    return( /*(+epsilon		fabs(epsilon) < 1.9E-7 */
		      (+0.39894228
		+tinv*(+0.01328592
		+tinv*(+0.00225319
		+tinv*(-0.00157565
		+tinv*(+0.00916281
		+tinv*(-0.02057706
		+tinv*(+0.02635537
		+tinv*(-0.01647633
		+tinv*(+0.00392377)))))))))/*)*/ * exp(x)/sqrt(x)
	    );
	};
}


/*
 *  Evaluate sinc(x) = sin(x)/x safely
 */
static double sinc(x)
double x;
{
	return(fabs(x) < 1E-50 ? 1.0 : sin(fmod(x,2*M_PI))/x);
}


/*
 *  Find number code for IIR-filter name
 */
static int lookup_iir_filtercode(name)
char *name;
{
	int i, j, code;
	char *p;

	/* g_ftyptab is in filtflp.c */
	i = 0;
	while(g_ftyptab[i].FTYPCODE != 0) {
	    j = 0;
	    p = g_ftyptab[i].FTYPABR;
	    while (p[j] != 0) {
		if ((p[j] | 0x20) != (name[j] | 0x20)) 
			break;
		j++;
	    }
	    if (p[j] == 0 && name[j] == 0)
		break;
	    i++;
	}
	return(g_ftyptab[i].FTYPCODE);
}

/*
 *  set_window_ifft
 *
 *  allocate a fft with a size which is the smallest 
 *  power of 2 greater or equal to nwdow
 */
static FFTDATA *set_window_ifft(nwdow)
int nwdow;
{
	int ldn, check;

	/* calculate ldn = ceil(log2(nwdow+1)) */
	ldn = -1; while ((nwdow >> 1) >> ++ldn) /* empty */ ;
	if ((1 << ldn) < nwdow)
		ldn++;
	/* (FFTDATA*)->n is exp2(ldn) */
	return(fft_init(ldn));
}

/*
 *  run_window_ifft
 *
 *  transform f-sampled window-spectrum to time domain,
 *  take window t-samples, discard trailing zeros
 */
static double run_window_ifft(wfftp, wdow, nwdow)
FFTDATA *wfftp;
double *wdow;
int nwdow;
{
	int i, nwfft = wfftp->n;
	double wnorm = 0.0;

	ifft(wfftp);
	/* if nwdow < nwfft we assume (wfftp->r)[i] to be zero for i >
	 * nwdow reason: we then had oversampled in frequency (to fit
	 * a nwfft of a power of 2) corresponding to some zeros in
	 * time (actually, values might be not really zero if function
	 * sampled in frequency wasn't sharply bandlimited (only
	 * chebyshev is)) */
	for (i = 0; i < nwdow; i++)
		wnorm += (wdow[i] *= (wfftp->r)[i]);
	fft_discard(wfftp);
	return(wnorm);
}

/*
 *  window (at)^(n-1)*exp(-at)
 */
static double tXexp(wdow, nwdow, nf3db, tshift, n)
double *wdow, nf3db;
int nwdow, n, tshift;
{
	double a, at, latn = 0.0, wnorm = 0.0;
	int i;

	/* For negative tshift discard -tshift samples, for positive
	 * tshift prepend tshift zeros-samples take a total of nwdow
	 * samples, then time-reverse */
	a = 2 * M_PI * nf3db / sqrt(exp(log((double)2)/n) - 1);
	for (i = 0; i < nwdow; i++) {
	    at = a * (nwdow-1-i-tshift);
	    if (tshift <= nwdow-1-i) {
		if (n > 1)
		    latn = (at > 0) ? log(at)*(n-1) : -1E30;
		wnorm += (wdow[i] *= exp(latn-at));
	    } else
		wdow[i] *= 0.0;
	}
	return(wnorm);
}

/*
 *  window_init
 *  ===========
 *
 *  Allocate an window array of size <nwdow>, initialize as `rectangle'
 */
double *window_init(nwdow)
int nwdow;
{
	int i;
	double *wdow;

	wdow = (double*)(malloc(sizeof(double)*nwdow));
	if (wdow == NULL)
		return(NULL);
	for (i = 0; i < nwdow; i++)
		wdow[i] = 1.0/nwdow;
	return(wdow);
}

/*
 *  window_discard
 *  ==============
 */
void window_discard(wdow)
double *wdow;
{
	free(wdow);
}

/*
 *  window_window
 *  =============
 *
 *  Calculate window of type <wtype> (known windows see window.h)
 *  and size <nwdow>, multiply it onto the array <wdow> and normalize
 *  to have a peak signal gain of <gain>. By chaining calls many windows
 *  may be multiplied on to each other (but only the last callers
 *  <gain> will be effective, previous gains will be normalized away).
 *  First use window_init() to allocate and initialize <wdow>.
 *  Return non-zero on error (return codes see window.h).
 *
 *  <sampling> determines sampling-mode (see window.h and discussion
 *  below). <alpha>, <nf3db>, <nfco> and <tshift> are window-dependend
 *  parameters (window/parameter-association see window.h).
 *
 *  Most window definitions as described by Harris, Proc. IEEE 66(1),
 *  51-83, 1978. Some windows by Nuttall, IEEE ASSP-29(1), 84-91,
 *  1981. Corrections on Harris' work by Nuttall are regarded. Own
 *  corrections on chebyshev and tukey windows. Modified-kaiser-bessel
 *  as described by Geckinly and Yavuz, IEEE ASSP-26, 501-507, 1978.
 */
int window_window(wdow, nwdow, wtype, gain, sampling, 
		  alpha, nf3db, nfco, tshift)
double *wdow, gain, alpha, nf3db, nfco;
int nwdow, sampling, tshift;
char *wtype;
{
	int i, m, o, code, nwfft, caught_mainlobe_flag;
	FFTDATA *wfftp;
	double wnorm, in, out, a, beta, c1, c2, c3, twid, A, B, C, y,
	  d[MAXD], g[MAXG], z[MAXZ];

	wnorm = 0.0;
	beta = M_PI * alpha; /* needed for kaiser-related windows */
	switch (sampling) {
	case WINDOW_PERIODIC:  m = nwdow;   break;
	case WINDOW_SYMMETRIC: m = nwdow-1; break;
	default:               return(WINDOW_ERROR_PRANGE);
	}
	if (m == 0)
		return(WINDOW_ERROR_PRANGE);
	/* 
	 *  On the use of window sampling ...
	 *  ---------------------------------
	 *
	 *  DFT-even/periodic/L=K  window sampling -> m=nwdow
	 *  finite/symmetric/L=K+1 window sampling -> m=nwdow-1
	 *
	 *  The concept of DFT-even/finite sampling has been
	 *  introduced by Harris [Proc. IEEE 66(1), 51-83, 1978]. It
	 *  addresses the question wether the endpoint of a sampled
	 *  time-limited symmetric window function should be part of
	 *  the coefficient vector (finite) or not (DFT-even). It is
	 *  somewhat artificial, as pointed out by Rohling and
	 *  Schuerman [Signal Processing 5, 127-138, North-Holland,
	 *  1983]. Yet the distinction still prevails, e.g. windows in
	 *  the the Signal Processing Toolbox for MATLAB have a
	 *  flag to choose sampling. I included both types. Here are
	 *  my own remarks on this distinction:
	 *
	 *  (1) Some windows (e.g. chebyshev) are DFT-even by
	 *  definition. To make them finite means forcing a zero as
	 *  last value, reducing space for significant values from
	 *  nwdow to nwdow-1. In these cases window sampling affects
	 *  only the definition of window size (L=K or L=K+1).
	 *
	 *  (2) Some windows are finite by definition and end in zeros
	 *  (e.g. hanning). To make them DFT-even means discarding
	 *  last value, enlarging space for other values from nwdow-1
	 *  to nwdow. This can be compensated for by reducing nwdow by
	 *  one. So again, this is just a matter of definition of
	 *  window size (L=K or L=K+1). Even more, the first value
	 *  could be discarded too, see SigProcToolBox MATLAB:
	 *  bartlett versus triangle. We never skip leading zeros in
	 *  this implementation.
	 *
	 *  (3) But for those windows finite by definition having
	 *  non-zero ending values (e.g. hamming) discarding the last
	 *  value changes the properties in a way that can not be
	 *  compensated by redefining window size. Yet there is no
	 *  such thing as a `correct' sampling, be it for DFT-analysis
	 *  or a FIR-filter-design: The sampling of the underlying
	 *  time-limited function (e.g. hamming: 0.56-0.44cos(2*pi*t),
	 *  0<=t<=T) causes aliasing in the frequency domain, which
	 *  can be safely reduced only be enlarging sample rate (means
	 *  enlarging window size).
	 *
	 *  (4) No general statement can be made which sampling is
	 *  more advantageous for a given window size (Rohling and
	 *  Schuerman). The difference of course vanishes with
	 *  increasing window size. Yet true optimality constraints
	 *  for a window can only be met on the basis a given window
	 *  size, not by sampling a time-continous window. A true
	 *  optimal window for example is the chebyshev window
	 *  (constraint: least sidelobe level). For general optimality
	 *  on windows see Adams [IEEE ASSP-39(8), 1753-1769, 1991].
	 *
	 *  (5) There are situations when sampling becomes important
	 *  even for a large window size: To meet perfect
	 *  reconstruction constraints in analysis/synthesis
	 *  filterbanks, for example in a DFT-overlap-add
	 *  system. Window sampling and block advance depend on each
	 *  other.
	 *
	 *  (6) For truncated causal (thus asymmetric) IIR-windows no
	 *  difference in sampling is implemented here, the reversed
	 *  time origin is always rightbound within the vector
	 *  wdow[0...nwdow-1]^T.
	 */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (a) fixed windows without parameters                             */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	if (!strcmp(wtype, "rectangle") ||
	    !strcmp(wtype, "boxcar") ||
	    !strcmp(wtype, "box-car") ||
	    !strcmp(wtype, "dirichlet")) {
	    for (i = 0; i < nwdow; i++) {
		wnorm += (wdow[i] *= 1.0);
	    }
	} else if (!strcmp(wtype, "triangle") ||
		   !strcmp(wtype, "fejer") ||
		   !strcmp(wtype, "bartlett") ||
		   !strcmp(wtype, "bartlet")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2.0*i)/m - 1.0;
		wnorm += (wdow[i] *= 1.0 - fabs(c1));
	    }
	} else if (!strcmp(wtype, "half-endpoints")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (i == 0 || i == m) ? 0.5 : 1.0;
		wnorm += (wdow[i] *= c1);
	    }
	} else if (!strcmp(wtype, "unraised-cosine") ||
		   !strcmp(wtype, "cosine-tip")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (M_PI*i)/m;
		wnorm += (wdow[i] *=  sin(c1));
	    }
	} else if (!strcmp(wtype, "hanning") ||
		   !strcmp(wtype, "hann") ||
		   !strcmp(wtype, "raised-cosine")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.5 - 0.5*cos(c1));
	    }
	} else if (!strcmp(wtype, "hamming")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.54 - 0.46*cos(c1));
	    }
	} else if (!strcmp(wtype, "exact-hamming")) { /* Nuttall81 */
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.53836 - 0.46164*cos(c1));
	    }
	} else if (!strcmp(wtype, "blackman")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.42 - 
				     0.50 * cos(c1) +
				     0.08 * cos(2*c1));
	    }
	} else if (!strcmp(wtype, "exact-blackman")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.42659071 - 
				     0.49656062 * cos(c1) +
				     0.07684867 * cos(2*c1));
	    }
	} else if (!strcmp(wtype, "blackman-harris-3-70-6")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.42323 - 
				     0.49755 * cos(c1) +
				     0.07922 * cos(2*c1));
	    }
	} else if (!strcmp(wtype, "blackman-harris-3-62-6")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.44959 - 
				     0.49364 * cos(c1) +
				     0.05677 * cos(2*c1));
	    }
	} else if (!strcmp(wtype, "blackman-harris-4-92-6")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.35875 - 
				     0.48829 * cos(c1) +
				     0.14128 * cos(2*c1) - 
				     0.01168 * cos(3*c1));
	    }
	} else if (!strcmp(wtype, "blackman-harris-4-74-6")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.40217 - 
				     0.49703 * cos(c1) +
				     0.09892 * cos(2*c1) - 
				     0.00188 * cos(3*c1));
	    }
	} else if (!strcmp(wtype, "nuttall-4-98-6")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.3635819 - 
				     0.4891775 * cos(c1) +
				     0.1365995 * cos(2*c1) - 
				     0.0106411 * cos(3*c1));
	    }
	} else if (!strcmp(wtype, "nuttall-4-93-18")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.355768 - 
				     0.487396 * cos(c1) +
				     0.144232 * cos(2*c1) - 
				     0.012604 * cos(3*c1));
	    }
	} else if (!strcmp(wtype, "nuttall-4-82-30")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.338946 - 
				     0.481973 * cos(c1) +
				     0.161054 * cos(2*c1) - 
				     0.018027 * cos(3*c1));
	    }
	} else if (!strcmp(wtype, "nuttall-4-60-42")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.312500 - 
				     0.468750 * cos(c1) +
				     0.187500 * cos(2*c1) - 
				     0.031250 * cos(3*c1));
	    }
	} else if (!strcmp(wtype, "nuttall-3-71-6")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.4243801 - 
				     0.4973406 * cos(c1) +
				     0.0782793 * cos(2*c1));
	    }
	} else if (!strcmp(wtype, "nuttall-3-64-18")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.40897 - 
				     0.50000 * cos(c1) +
				     0.09103 * cos(2*c1));
	    }
	} else if (!strcmp(wtype, "nuttall-3-46-30")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2*M_PI*i)/m;
		wnorm += (wdow[i] *= 0.375 - 
				     0.500 * cos(c1) +
				     0.125 * cos(2*c1));
	    }
	} else if (!strcmp(wtype, "riemann") ||
		   !strcmp(wtype, "sinc-mainlobe")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = 2 * M_PI * (i - m/2.0) / m;
		wnorm += (wdow[i] *= sinc(c1));
	    }
	} else if (!strcmp(wtype, "bohman") ||
		   !strcmp(wtype, "papoulis")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = fabs((2.0*i)/m - 1.0);
		c2 = (1.0 - c1)*(cos(M_PI*c1)) + sin(M_PI*c1)/M_PI;
		wnorm += (wdow[i] *= c2);
	    }
	} else if (!strcmp(wtype, "riesz") ||
		/* !strcmp(wtype, "parzen") || ambigious within Harris78 */
		   !strcmp(wtype, "bochner")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2.0*i)/m - 1.0;
		c2 = 1.0 - c1*c1;
		wnorm += (wdow[i] *= c2);
	    }
	} else if (!strcmp(wtype, "de-la-valle-poussin") ||
		   !strcmp(wtype, "parzen") ||  /* Geckinly78 */
		   !strcmp(wtype, "jackson")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = fabs((2.0*i)/m - 1.0);
		if (c1 <= 0.5) {
		     c2 = 1.0 - 6.0 * (c1*c1) * (1.0 - c1);
		} else {
		     c2 = 2.0 * (1.0 - c1) * (1.0 - c1) * (1.0 - c1);
		}
		wnorm += (wdow[i] *= c2);
	    }


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (b) windows taking parameter alpha                               */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	} else if (!strcmp(wtype, "cos")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (M_PI*i)/m;
		c2 = sin(c1);
		c3 = (c2 > 0 ? exp(alpha * log(c2)) : 0.0);
		wnorm += (wdow[i] *= c3);
	    }
	} else if (!strcmp(wtype, "tukey") ||
		   !strcmp(wtype, "cosine-taper") ||
		   !strcmp(wtype, "raised-cosine-taper")) {
	    if (alpha < 0.0 || alpha > 1.0)
		return(WINDOW_ERROR_PRANGE);
	    for (i = 0; i < nwdow; i++) {
		/* Note: Eq.(38) in Harris78 has 3 errors:
		   n -> |n|, alpha -> (1- alpha), pi -> 2*pi */
		c1 = (1.0 - (1.0 - alpha)) * m / 2.0;
		c2 = (1.0 + (1.0 - alpha)) * m / 2.0;
		if (i > c2) {
		    c3 =  0.5 + 0.5 * cos(M_PI * (i - c2) / c1);
		    wnorm += (wdow[i] *= c3);
		} else if (i < c1) {
		    c3 =  0.5 + 0.5 * cos(M_PI * (i - c1) / c1);
		    wnorm += (wdow[i] *= c3);
		} else
		    wnorm += (wdow[i] /* *= 1.0 */);
	    }
	} else if (!strcmp(wtype, "unraised-cosine-taper")) {
	    if (alpha < 0.0 || alpha > 1.0)
		return(WINDOW_ERROR_PRANGE);
	    for (i = 0; i < nwdow; i++) {
		c1 = (1.0 - (1.0 - alpha)) * m / 2.0;
		c2 = (1.0 + (1.0 - alpha)) * m / 2.0;
		if (i > c2) {
		    c3 =  cos(M_PI * (i - c2) / (2 * c1));
		    wnorm += (wdow[i] *= c3);
		} else if (i < c1) {
		    c3 =  cos(M_PI * (i - c1) / (2 * c1));
		    wnorm += (wdow[i] *= c3);
		} else
		    wnorm += (wdow[i] /* *= 1.0 */);
	    }
	} else if (!strcmp(wtype, "poisson")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2.0*i)/m - 1.0;
		wnorm += (wdow[i] *= exp(-alpha*fabs(c1)));
	    }
	} else if (!strcmp(wtype, "cauchy") ||
		/* !strcmp(wtype, "poisson") || ambigious within Harris78 */
		   !strcmp(wtype, "abel")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = (2.0*i)/m - 1.0;
		wnorm += (wdow[i] *= 1.0/(1.0 + (alpha*alpha*c1*c1)));
	    }

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (c) kaiser-related windows taking parameter beta                 */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	} else if (!strcmp(wtype, "kaiser-bessel") ||
		   !strcmp(wtype, "kaiser")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = 2.0 * (i - m/2.0) / m;
		c2 = I0(beta * sqrt(1.0 - c1 * c1));
		wnorm += (wdow[i] *= c2);
	    }
	} else if (!strcmp(wtype, "modified-kaiser-bessel") ||
		   !strcmp(wtype, "modified-kaiser")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = 2.0 * (i - m/2.0) / m;
		c2 = I0(beta * sqrt(1.0 - c1 * c1));
		wnorm += (wdow[i] *= c2 - 1);
	    }

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (d) truncated acausal-IIR-windows taking parameters nf3db, nfco  */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	} else if (!strcmp(wtype, "sinc")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = 2 * M_PI * nfco * (i - m/2.0);
		wnorm += (wdow[i] *= sinc(c1));
	    }
	} else if (!strcmp(wtype, "gaussian") ||
		   !strcmp(wtype, "weierstrass")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = alpha * (i - m/2.0) / m;
		wnorm += (wdow[i] *= exp(-2*c1*c1));
	    }
	} else if (!strcmp(wtype, "gaussian-f3db")) {
	    for (i = 0; i < nwdow; i++) {
		c1 = M_PI * nf3db * (i - m/2.0);
		wnorm += (wdow[i] *= exp(-2*c1*c1/(log(2.0))));
	    }

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (e) truncated causal-IIR-windows from pole plans via bilinear-   */
/*	    transform taking parameters nf3db, tshift                    */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	} else if ((code = lookup_iir_filtercode(wtype)) != 0) {
	    /* make parameters, reset state, run impulse through
	     * filter for negative tshift discard -tshift samples, for
	     * positive tshift prepend tshift zeros-samples take a
	     * total of nwdow samples, then time-reverse */
	    filterparms(code, (double)1.0,
		    nf3db, (double)0, (double)0, d, g);
	    for (i = 0; i < MAXZ; i++)
		z[i] = 0.0;
	    in = 1.0; /* discard -tshift samples if tshift < 0 */
	    for (i = -tshift; i > 0; i--) {
		filtering(code, in, z, d, g, &out);
		in = 0.0;
	    }
	    for (i = nwdow-1; i >= 0; i--) {
	        if (tshift == nwdow-1-i)
		    in = 1.0;
		else
		    in = 0.0;
		filtering(code, in, z, d, g, &out);
		wnorm += (wdow[i] *= out);
	    }

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (f) truncated causal-IIR-Windows from pole-plans via sampled     */
/*	    impuls response taking parameters nf3db, tshift              */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	} else if (!strcmp(wtype, "exp") ||
		   !strcmp(wtype, "exponential") ||
		   !strcmp(wtype, "t0exp")) {
	    wnorm = tXexp (wdow, nwdow, nf3db, tshift, 1);
	} else if (!strcmp(wtype, "t1exp")) {
	    wnorm = tXexp (wdow, nwdow, nf3db, tshift, 2);
	} else if (!strcmp(wtype, "t2exp")) {
	    wnorm = tXexp (wdow, nwdow, nf3db, tshift, 3);
	} else if (!strcmp(wtype, "t3exp")) {
	    wnorm = tXexp (wdow, nwdow, nf3db, tshift, 4);


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (g) Windows from aliasing-free frequency domain specification	 */
/*	    sample f-domain, IDFT, t-domain is time-limited -> no alias. */
/*	    Parameter is alpha.                                          */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	} else if (!strcmp(wtype, "chebyshev") ||
		   !strcmp(wtype, "dolph-chebyshev")) {
	    /* Note: The dolph-chebyshev specification presented by
	     * Harris78 is incorrect because the order `o' (degree) of
	     * the chebyshev-polynominal must be smaller than window
	     * size (see Diderich, Proc. IEEE 62, 1395-1396, Oct74).
	     * In the case of f-oversampling (meaning o+1 < nwfft) the
	     * time-forward shift of o/2 is important, because only
	     * the first nwdow samples of the inverse DFT are taken
	     * (nwdow=(m=o+1) in reasonable cases, otherwise have
	     * nwdow-m trailing zeros). In theory the shift should be
	     * (o+1)/2 (see Diderich74) not o/2, which doesn't work in
	     * practice (in this code?) --- HELP!!! */
	    if ((wfftp = set_window_ifft(nwdow)) == NULL)
		return(WINDOW_ERROR_NOMEM);
	    nwfft = wfftp->n; /* f-samples of chebyshev-polynominal (CP) */
	    o = m - 1;	      /* order of CP, have o+1 nonzero t-samples */
	    twid = -o * M_PI / nwfft; /* do forward-shift o/2 t-samples  */ 
	    beta = (o > 0 ?
		cosh(acosh(exp(M_LN10*alpha))/o) : 0.0) ;
	    for (i = -nwfft/2; i < (nwfft+1)/2; i++) {
		c1 = i * M_PI / nwfft;
		c2 = beta * cos(c1);
		c3 = (fabs(c2) <= 1.0 ?
		      cos(o * acos(c2)) : cosh(o * acosh(c2)));
		(wfftp->r)[(i+nwfft)%nwfft] = c3*cos(i*twid);
		(wfftp->i)[(i+nwfft)%nwfft] = c3*sin(i*twid);
	    }
	    wnorm = run_window_ifft(wfftp, wdow, nwdow);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      (h) Windows from aliasing-prone frequency domain specification   */
/*	    sample f-domain, IDFT, result is not time-limited -> alias.  */
/*                              USE WITH CARE                            */
/*	    Parameter is alpha (for kaiser-bessel-fds: beta)             */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	} else if (!strcmp(wtype, "van-der-maas")) { /* Nuttall81 */
	    /* van-der-maas is almost chebyshev, has impulses at
	     * both ends, that's why similar f-sampling and time-
	     * shifting is done, but f-oversampling now >= 2 */
	    if ((wfftp = set_window_ifft(2*nwdow)) == NULL)
		return(WINDOW_ERROR_NOMEM);
	    nwfft = wfftp->n; /* f-samples of cos(sqrt((piLf)^2+B^2)) */
	    o = m - 1;	      /* we get o+1 nonzero t-samples */
	    twid = -o * M_PI / nwfft;  /* forward-shift o/2 t-samples */
	    B = acosh(exp(M_LN10*alpha));
	    for (i = -nwfft/2; i < (nwfft+1)/2; i++) {
		c1 = i * (o * M_PI) / nwfft ;
		c2 = c1*c1-B*B;
		c3 = (c2 >= 0 ? 
		      cos(sqrt(c2)) : cosh(sqrt(-c2)));
		(wfftp->r)[(i+nwfft)%nwfft] = c3*cos(i*twid);
		(wfftp->i)[(i+nwfft)%nwfft] = c3*sin(i*twid);
	    }
	    wnorm = run_window_ifft(wfftp, wdow, nwdow);

	} else if (!strcmp(wtype, "barcilon-temes")) {
	    /* similar to kaiser, truncation in time domain less
	     * critical than chebyshev/van-der-maas, f-oversmp >=2
	     * Performance is basically as claimed by Harris78, though
	     * effective alpha seems to deviate slightly */
	    if ((wfftp = set_window_ifft(2*nwdow)) == NULL)
		return(WINDOW_ERROR_NOMEM);
	    nwfft = wfftp->n;          /* number of f-samples */
	    twid = -m * M_PI / nwfft;  /* forward-shift m/2 t-samples */
	    A = sqrt(exp(M_LN10*2*alpha)-1);
	    B = exp(M_LN10*alpha);
	    C = acosh(B);
	    beta = cosh(C/m);
	    caught_mainlobe_flag = 0;
	    for (i = -nwfft/2; i < (nwfft+1)/2; i++) {
		c1 = beta * cos(i * M_PI / nwfft);
		if (fabs(c1) <= 1.0) {
		    y  = m * acos(c1);
		    c2 = A*cos(y) + B*(sin(y)*y/C);
		    c3 = c2 / ((C+A*B)*((y*y)/(C*C) + 1.0));
		} else {
		    y  = m * acosh(c1);
		    if (fabs(C-y) > 1E-5) { /* catch mainlobe center */ 
		        /* numerically unstable ... */
			c2 = A*cosh(y) - B*(sinh(y)*y/C);
			c3 = c2 / ((C+A*B)*(1.0 - (y*y)/(C*C)));
		    } else {
			/* should apply de l'hospital, 0.5 is empirical */
			c3 = 0.5;
			/* we should be here only once, otherwise
			   there'd be a numerical problem */
			if (caught_mainlobe_flag)
			    return(WINDOW_ERROR_NUMERIC);
			caught_mainlobe_flag = -1;
		    }
		}
		(wfftp->r)[(i+nwfft)%nwfft] = c3*cos(i*twid);
		(wfftp->i)[(i+nwfft)%nwfft] = c3*sin(i*twid);
	    }
	    wnorm = run_window_ifft(wfftp, wdow, nwdow);

	} else if (!strcmp(wtype, "kaiser-fds") ||
		   !strcmp(wtype, "kaiser-bessel-fds")) {
	    /* PREFER TIME-DOMAIN SPEC `kaiser-bessel' !!!
	       f-oversampling >= 2x */
	    if ((wfftp = set_window_ifft(nwdow*2)) == NULL)
		return(WINDOW_ERROR_NOMEM);
	    nwfft = wfftp->n;          /* number of f-samples           */
	    twid = -m * M_PI / nwfft;  /* forward-shift m/2 t-samples	*/
	    for (i = -nwfft/2; i < (nwfft+1)/2; i++) {
		c1 = i * (m * M_PI) / nwfft ;
		c2 = c1*c1-beta*beta;
		c3 = (fabs(c2) > 1E-30) ?
			(c2 >= 0 ? 
				sin(sqrt(c2))/sqrt(c2) : 
				sinh(sqrt(-c2))/sqrt(-c2)
			) : 0.0;
		(wfftp->r)[(i+nwfft)%nwfft] = c3*cos(i*twid);
		(wfftp->i)[(i+nwfft)%nwfft] = c3*sin(i*twid);
	    }
	    wnorm = run_window_ifft(wfftp, wdow, nwdow);
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*      END OF WINDOW IMPLEMENTATION                                     */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	} else
	    return(WINDOW_ERROR_NOWIN); /* wtype/window unknown */

	if (fabs(wnorm) < 1E-30)	/* window has vanishing DC-gain */
	    return(WINDOW_ERROR_ZERODC);

	a = gain / wnorm;               /* normalize and set gain */
	for (i = 0; i < nwdow; i++)
		wdow[i] *= a;

	return(WINDOW_OK);
}

/*
 *  window_raise
 *  ============
 *
 *  Normalize window <wdow> of size <nwdow> to maximum one
 *  (previous gain thus will be lost); raise each coefficent
 *  by formula
 *
 *	c = raise + (1 - raise) * c 
 *
 *  and normalize resulting window to DC-gain <gain>.
 *  Return WINDOW_OK or WINDOW_ERROR_ZERODC.
 */
int window_raise(wdow, nwdow, raise, gain)
double *wdow, raise, gain;
int nwdow;
{
	int i;
	double a, max, wnorm;

	max = -1E+30;
	for (i = 0; i < nwdow; i++)	/* find maximum value */
		max = wdow[i] > max ? wdow[i] : max;

	wnorm = 0.0;
	for (i = 0; i < nwdow; i++)
	    wnorm += (wdow[i] = raise + (1.0 - raise) * wdow[i] / max);

	if (fabs(wnorm) < 1E-30)	/* result has vanishing DC-gain */
	    return(WINDOW_ERROR_ZERODC);
	a = gain / wnorm;		/* normalize and set gain */
	for (i = 0; i < nwdow; i++)
		wdow[i] *= a;

	return(WINDOW_OK);
}
/*
 * EOT
 *
 ************************************************************************/
