/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/*************************************************************************
 *
 *   Name    : window.h
 *   Version : 1999/08/20 1.15
 *   Comment : Data windows for FIR-design and signal analysis
 *   Author  : Markus Mummert
 *   History : (see .c-file)
 *
 */
#ifndef WINDOW_INCLUDED	   /* multiple-include protection */
#define WINDOW_INCLUDED

extern double *window_init(/* int nwdow */); 
/* ... allocate window of size <nwdow>, initalize as `rectangle',
       return pointer to window or NULL on out-of-memory */
extern int   window_window(/* double *wdow, int nwdow,    char *wtype,
			      double gain,  int sampling, double alpha,
			      double nf3db, double nfco,  int tshift */);
/* ... (this call is optional) multiply window <wtype> over exisiting
       window, normalize result for DC-gain <gain> (more parameters and
       return codes see below) */
extern int    window_raise(/* double *wdow, int nwdow, 
			      double raise, double gain*/);
/* ... (this call is optional) normalize window <wtype> to maximum one 
       (thus lose previous gain), apply window=raise+(1-raise)*window,
       normalize to <gain>, return WINDOW_OK or WINDOW_ERROR_ZERODC */
extern void window_discard(/* double *wdow */);
/* ... free allocated window space */

/*
 *  Known identifiers for <wtype>
 */
#define WINDOW_LIST "\
2p1 3p1 4p1 8p1 b2 b3 b4 b8 \
barcilon-temes blackman \
blackman-harris-3-62-6 blackman-harris-3-70-6 \
blackman-harris-4-74-6 blackman-harris-4-92-6 \
bohman cauchy chebyshev cos de-la-valle-poussin \
exp exact-blackman exact-hamming gaussian gaussian-f3db \
half-endpoints hamming hanning \
kaiser-bessel kaiser-bessel-fds modified-kaiser-bessel \
nuttall-3-46-30 nuttall-3-64-18 nuttall-3-71-6 \
nuttall-4-60-42 nuttall-4-82-30 nuttall-4-93-18 nuttall-4-98-6 \
p1 p2 p3 p4 p8 pg2 pg3 pg4 pg8 poisson \
rectangle riemann riesz sinc \
t1exp t2exp t3exp triangle tukey \
unraised-cosine unraised-cosine-taper van-der-maas \
"

#define WINDOW_ALIAS_LIST "\
abel=cauchy \
bartlett=triangle bartlet=triangle bochner=riesz \
box-car=rectangle boxcar=rectangle \
cosine-taper=tukey cosine-tip=unraised-cosine \
dirichlet=rectangle dolph-chebyshev=chebyshev \
exponential=exp fejer=triangle \
hann=hanning jackson=de-la-valle-poussin \
kaiser=kaiser-bessel kaiser-fds=kaiser-bessel-fds \
modified-kaiser=modified-kaiser-bessel \
parzen=de-la-valle-poussin papoulis=bohman \
raised-cosine=hanning raised-cosine-taper=tukey \
sinc-mainlobe=riemann t0exp=exp weierstrass=gaussian \
"

/*
 *  Return codes
 */
#define WINDOW_OK            0  /* call succeeded */
#define WINDOW_ERROR_NOMEM   1  /* not enough memory */
#define WINDOW_ERROR_NOWIN   2  /* unknown window */
#define WINDOW_ERROR_NOTPOW2 3  /* nwdow is not power of 2 */
#define WINDOW_ERROR_ZERODC  4	/* window has no dc (due to tshift?) */
#define WINDOW_ERROR_NUMERIC 5  /* numerical error */
#define WINDOW_ERROR_PRANGE  10 /* parameter out of range */

/* 
 *  <sampling> determines wether endpoint of a sampled time-limited
 *  symmetric window function is part of the window sample vector
 *  (symmetric/finite/nwdow=m+1) or not (periodic/DFT-even/nwdow=m).
 *  chebyshev is `periodic' by definition, making it `symmetric' means
 *  having a zero as endpoint which is uncommon practice. For perfect
 *  reconstruction constraints in overlap-add-filterbanks `periodic'
 *  likely is the right choice. In all other cases use `symmetric' as
 *  default -- the difference vanishes with increasing window size.
 *  (Also see elaborate discussion in the source file.)
 */
#define WINDOW_SYMMETRIC  1001  /* sampling nwdow=m+1 (finite, Harris78) */
#define WINDOW_PERIODIC   100   /* sampling nwdow=m (DFT-even, Harris78) */

/*  
 *  Window-types and associated parameters. Most windows are
 *  described by Harris [Proc. IEEE 66(1), p51-83, 1978] or by Nuttall
 *  [IEEE ASSP-29(1), p84-91, 1981] (partly correcting Harris). 
 *  modified-kaiser-bessel as described by Geckinly and Yavuz
 *  [IEEE ASSP-26, 501-507, 1978]. Window/filter design with
 *  kaiser-bessel see Oppenheim and Schafer [`Discrete Time-Signal
 *  Processing', Prentice-Hall, 2nd ed., 1999].
 *  
 *  (a)	Windows without parameters:
 *
 *		blackman-harris-3-62-6
 *		blackman-harris-3-70-6
 *		blackman-harris-4-92-6
 *		blackman-harris-4-74-6
 *		blackman
 *		bohman			aka papoulis
 *		de-la-valle-poussin	aka jackson, parzen
 *		exact-blackman
 *		exact-hamming		Nuttall81
 *		half-endpoints		take half of first and last coeff.
 *		hamming
 *		hanning
 *		nuttall-4-98-6
 *		nuttall-4-93-18
 *		nuttall-4-82-30
 *		nuttall-4-60-42
 *		nuttall-3-71-6
 *		nuttall-3-64-18
 *		nuttall-3-46-30
 *		poisson
 *		rectangle
 *		riemann			sinc-mainlobe
 *		riesz
 *		triangle
 *		unraised-cosine
 *
 *  (b)	Windows, parameter is <alpha> (as defined in Harris78)
 *
 *		cauchy
 *		cos
 *		tukey			Eq.(38) in Harris78 is incorrect
 *		unraised-cosine-taper	as tukey, but unraised-cosine
 *
 *  (c)	Windows, parameter is <alpha> (as defined in Harris78, common
 *      parameter in literature also is alpha' = beta = B = pi * alpha):
 *
 *		kaiser-bessel
 *		modified-kaiser-bessel
 *
 *  (d) Non-causal IIR-windows, truncated both sides to match <nwdow>,
 *	paramter is alpha (as defined in Harris78) for gaussian,
 *      normalized 3dB-cutoff frequency <nf3db> (nf3db=0.2650*alpha/nwdow) 
 *	for gaussian-f3db and normalized cutoff-frequency <nfco> for sinc.
 *
 *		gaussian	c*exp(-x*x/2),x=alpha*(2*i-m)/2
 *		gaussian-f3db	c*exp(-x*x/log(2)/2),x=pi*(2*i-m)*nf3db
 *		sinc		c*sin(x)/x, x=pi*(2*i-m)*nfco
 *
 *      Run i from 0 to nwdow-1, m see <sampling>.
 *				
 *  (e)	Causal-IIR-windows, constructed from time-continuous lowpasses
 *	of normalized 3dB-cutoff frequency <nf3db>. Start sampling at
 *	at zero-time, offset by <tshift> samples, read <nwdow> samples,
 *	then time-reverse samples.
 *
 *		exp		single real pole
 *		tXexp		multiple real pole of order X+1, X=1,2,3
 *
 *  (f)	Causal-IIR-windows, constructed from time-continuous lowpass pole
 *	plans by bilinear-transform, using normalized 3dB-cutoff frequency
 *	<nf3db> as fixpoint. Start reading impuls-response of time-discrete 
 *	realization at zero-time offset by <tshift> samples, read <nwdow> 
 *	samples, then time-reverse samples.
 *
 *		bN		Thomson/Bessel of order N=2,3,4,8
 *		Np1		multiple real pole, order N=2,3,4,8
 *		p1		single real pole
 *		pN		Butterworth, order N=2,3,4,8
 *		pgN		Gaussian-approximation, order N=2,3,4,8
 *				
 *	Coefficients for pgN by M.Dishal [Electrical Communication 36(1),
 *	p.3-26, 1959], all others standard.
 *
 *  (g)	Windows generated from aliasing-free frequency domain specification, 
 *	parameter is <alpha> (as defined in Harris) specifying the
 *	main_to_sidelobe_level/20dB:
 *
 *		chebyshev	Eq.(45) in Harris78 is incorrect, compare
 *				Diderich, Proc. IEEE 62, 1395-1396, Oct'74
 *
 * (h)  Windows from aliasing-prone frequency domain specification,
 *	parameter is <alpha> (as defined in Harris78). PREFERABLY USE
 *	chebyshev for van-der-maas OR kaiser-bessel for kaiser-bessel-fds.
 *	Performance of barcilon-temes is basically as claimed by Harris78,
 *	though effective alpha seems to deviate a little.
 *
 *		kaiser-bessel-fds   beta  = pi * alpha
 *		van-der-maas	    alpha = main_to_sidelobe_level/20dB
 *		barcilon-temes	    effective alpha deviates slightly
 * */

#endif /* WINDOW_INCLUDED */
/*
 * EOT
 *
 ************************************************************************/
